# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""FederatedData interface for providing access to a federated dataset."""

import abc
from typing import Any, Callable, Iterable, Iterator, Optional, Tuple

from fedjax.core import client_datasets
import numpy as np

# A client id is simply some binary bytes.
ClientId = bytes


class ClientPreprocessor:
  """A chain of preprocessing functions on all examples of a client dataset.

  This is very similar to :class:`fedjax.BatchPreprocessor`, with the main
  difference being that ClientPreprocessor also takes ``client_id`` as input.

  See the discussion in :class:`fedjax.BatchPreprocessor` regarding when to
  use which.
  """

  def __init__(self,
               fns: Iterable[Callable[[ClientId, client_datasets.Examples],
                                      client_datasets.Examples]] = ()):
    self._fns = tuple(fns)

  def __call__(self, client_id: ClientId,
               examples: client_datasets.Examples) -> client_datasets.Examples:
    if not self._fns:
      return examples
    # Make a copy to guard against fns that modify their input.
    out = dict(examples)
    for f in self._fns:
      out = f(client_id, out)
    client_datasets.assert_consistent_rows(out)
    return out

  def append(
      self, fn: Callable[[ClientId, client_datasets.Examples],
                         client_datasets.Examples]
  ) -> 'ClientPreprocessor':
    """Creates a new ClientPreprocessor with fn added to the end."""
    return ClientPreprocessor(self._fns + (fn,))

  def __str__(self) -> str:
    return f'ClientPreprocessor({self._fns})'

  def __repr__(self) -> str:
    return str(self)


# A common default preprocessor that does nothing.
NoOpClientPreprocessor = ClientPreprocessor()


class FederatedData(abc.ABC):
  """FederatedData interface for providing access to a federated dataset.

  A FederatedData object serves as a mapping from client ids to client
  datasets and client metadata.

  **Access methods with better I/O efficiency**

  For large federated datasets, it is not feasible to load all client datasets
  into memory at once (whereas loading a single client dataset is assumed to be
  feasible). Different implementations exist for different on disk storage
  formats. Since sequential read is much faster than random read for most
  storage technologies, FederatedData provides two types of methods for
  accessing client datasets,

  1.  :meth:`clients` and :meth:`shuffled_clients` are sequential read
      friendly, and thus recommended whenever appropriate.
  2.  :meth:`get_clients` requires random read, but prefetching is possible.
      This should be preferred over :meth:`get_client`.
  3.  :meth:`get_client` is usually the slowest way of accessing client
      datasets, and is mostly intended for interactive exploration of a small
      number of clients.

  **Preprocessing**

  :class:`~fedjax.ClientDataset` produced by FederatedData can hold a
  :class:`~fedjax.BatchPreprocessor`, customizable via :meth:`preprocess_batch`.
  Additionally, another "client" level :class:`ClientPreprocessor`, customizable
  via :meth:`preprocess_client`, can be used to apply transformations on
  examples from the entire client dataset before a
  :class:`~fedjax.ClientDataset` is constructed.
  """

  @abc.abstractmethod
  def slice(self,
            start: Optional[ClientId] = None,
            stop: Optional[ClientId] = None) -> 'FederatedData':
    """Returns a new FederatedData restricted to client ids in the given range.

    The returned FederatedData includes clients whose ids are,

    - Greater than or equal to ``start`` when ``start`` is not None;
    - Less than ``stop`` when ``stop`` is not None.

    Args:
      start: Start of client id range.
      stop: Stop of client id range.

    Returns:
      FederatedData.
    """

  # Client metadata access.

  @abc.abstractmethod
  def num_clients(self) -> int:
    """Returns the number of clients.

    If it is too expensive or otherwise impossible to obtain the result, an
    implementation may raise an exception.
    """

  @abc.abstractmethod
  def client_ids(self) -> Iterator[ClientId]:
    """Returns an iterator of client ids as bytes.

    There is no requirement on the order of iteration.
    """

  @abc.abstractmethod
  def client_sizes(self) -> Iterator[Tuple[ClientId, int]]:
    """Returns an iterator of all (client id, client size) pairs.

    This is often more efficient than making multiple :meth:`client_size` calls.
    There is no requirement on the order of iteration.
    """

  @abc.abstractmethod
  def client_size(self, client_id: ClientId) -> int:
    """Returns the number of examples in a client dataset."""

  # Client access.

  @abc.abstractmethod
  def clients(self) -> Iterator[Tuple[ClientId, client_datasets.ClientDataset]]:
    """Iterates over clients in a deterministic order.

    Implementation can choose whatever order that makes iteration efficient.
    """

  @abc.abstractmethod
  def shuffled_clients(
      self,
      buffer_size: int,
      seed: Optional[int] = None
  ) -> Iterator[Tuple[ClientId, client_datasets.ClientDataset]]:
    """Iterates over clients with a repeated buffered shuffling.

    Shuffling should use a buffer size of at least ``buffer_size`` clients.
    The iteration should repeat forever, with usually a different order in each
    pass.

    Args:
      buffer_size: Buffer size for shuffling.
      seed: Optional random number generator seed.

    Returns:
      Iterator.
    """

  @abc.abstractmethod
  def get_clients(
      self, client_ids: Iterable[ClientId]
  ) -> Iterator[Tuple[ClientId, client_datasets.ClientDataset]]:
    """Gets multiple clients in order with one call.

    Clients are returned in the order of ``client_ids``.

    Args:
      client_ids: Client ids to load.

    Returns:
      Iterator.
    """

  @abc.abstractmethod
  def get_client(self, client_id: ClientId) -> client_datasets.ClientDataset:
    """Gets one single client dataset.

    Prefer :meth:`clients`, :meth:`shuffled_clients`, or :meth:`get_clients`
    when possible.

    Args:
      client_id: Client id to load.

    Returns:
      The corresponding ClientDataset.
    """

  # Preprocessing.

  @abc.abstractmethod
  def preprocess_client(
      self, fn: Callable[[ClientId, client_datasets.Examples],
                         client_datasets.Examples]
  ) -> 'FederatedData':
    """Registers a preprocessing function to be called on all examples of a client before passing them to construct a ClientDataset."""

  @abc.abstractmethod
  def preprocess_batch(
      self, fn: Callable[[client_datasets.Examples], client_datasets.Examples]
  ) -> 'FederatedData':
    """Registers a preprocessing function to be called after batching in ClientDatasets."""


# Functions for treating a federated dataset as a single centralized dataset.


def shuffle_repeat_batch_federated_data(
    fd: FederatedData,
    batch_size: int,
    client_buffer_size: int,
    example_buffer_size: int,
    seed: Optional[int] = None) -> Iterator[client_datasets.Examples]:
  """Shuffle-repeat-batch all client datasets in a federated dataset for training a centralized baseline.

  Shuffling is done using two levels of buffered shuffling, first at the client
  level, then at the example level.

  This produces an infinite stream of batches. itertools.islice() can be used to
  cap the number of batches, if so desired.

  Args:
    fd: Federated dataset.
    batch_size: Desired batch size.
    client_buffer_size: Buffer size for client level shuffling.
    example_buffer_size: Buffer size for example level shuffling.
    seed: Optional RNG seed.

  Yields:
    Batches of preprocessed examples.
  """
  rng = np.random.RandomState(seed)
  datasets = (client_dataset for _, client_dataset in fd.shuffled_clients(
      client_buffer_size, rng.randint(1 << 32)))
  yield from client_datasets.buffered_shuffle_batch_client_datasets(
      datasets, batch_size=batch_size, buffer_size=example_buffer_size, rng=rng)


def padded_batch_federated_data(fd: FederatedData,
                                hparams: Optional[
                                    client_datasets.PaddedBatchHParams] = None,
                                **kwargs) -> Iterator[client_datasets.Examples]:
  """Padded batch all client datasets, useful for evaluation on the entire federated dataset.

  Args:
    fd: Federated dataset.
    hparams: See :func:`fedjax.padded_batch_client_datasets`.
    **kwargs: See :func:`fedjax.padded_batch_client_datasets`.

  Yields:
    Batches of preprocessed examples.
  """
  datasets = (client_dataset for _, client_dataset in fd.clients())
  yield from client_datasets.padded_batch_client_datasets(
      datasets, hparams, **kwargs)


def intersect_slice_ranges(
    current_start: Optional[ClientId], current_stop: Optional[ClientId],
    new_start: Optional[ClientId], new_stop: Optional[ClientId]
) -> Tuple[Optional[ClientId], Optional[ClientId]]:
  """Intersects the current slice range and the new slice range.

  This is a helper function for FederatedData implementations for ensuring
  slicing does not enlarge the range of client ids.

  Args:
    current_start: Current start of the slice range.
    current_stop: Current stop of the slice range.
    new_start: New start of the slice range.
    new_stop: New stop of the slice range.

  Returns:
    Normalized slice range that is the intersection of the two input ranges.
  """
  if current_start is not None:
    if new_start is None:
      new_start = current_start
    else:
      new_start = max(current_start, new_start)
  if current_stop is not None:
    if new_stop is None:
      new_stop = current_stop
    else:
      new_stop = min(current_stop, new_stop)
  return new_start, new_stop


class RepeatableIterator:
  """Repeats a base iterable after the end of the first pass of iteration.

  Because this is a stateful object, it is not thread safe, and all usual
  caveats about accessing the same iterator at different locations apply. For
  example, if we make two map calls to the same RepeatableIterator, we must make
  sure we do not interleave `next()` calls on these. For example, the following
  is safe because we finish iterating on m1 before starting to iterate on m2., ::

    it = RepeatableIterator(range(4))
    m1 = map(lambda x: x + 1, it)
    m2 = map(lambda x: x * x, it)
    # We finish iterating on m1 before starting to iterate on m2.
    print(list(m1), list(m2))
    # [1, 2, 3, 4] [0, 1, 4, 9]

  Whereas interleaved access leads to confusing results, ::

    it = RepeatableIterator(range(4))
    m1 = map(lambda x: x + 1, it)
    m2 = map(lambda x: x * x, it)
    print(next(m1), next(m2))
    # 1 1
    print(next(m1), next(m2))
    # 3 9
    print(next(m1), next(m2))
    # StopIteration!

  In the first pass of iteration, values fetched from the base iterator will be
  copied onto an internal buffer (except for a few builtin containers where
  copying is unnecessary). When each pass of iteration finishes (i.e. when
  __next__() raises StopIteration), the iterator resets itself to the start of
  the buffer, thus allowing a subsequent pass of repeated iteration.

  In most cases, if repeated iterations are required, it is sufficient to simply
  copy values from an iterator into a list. However, sometimes an iterator
  produces values via potentially expensive I/O operations (e.g. loading client
  datasets), RepeatableIterator can interleave I/O and JAX compute to decrease
  accelerator idle time in this case.
  """

  def __init__(self, base: Iterable[Any]):
    if any(
        isinstance(base, container)
        for container in (list, tuple, dict, str, bytes)):
      # No copying for builtin containers that are already repeatable.
      self._first_pass = False
      self._iter = iter(base)
      self._buf = base
    else:
      # General case, copying required.
      self._first_pass = True
      self._iter = iter(base)
      self._buf = []

  def __iter__(self) -> Iterator[Any]:
    return self

  def __next__(self) -> Any:
    try:
      value = next(self._iter)
    except StopIteration:
      if self._first_pass:
        self._first_pass = False
      self._iter = iter(self._buf)
      raise
    if self._first_pass:
      self._buf.append(value)
    return value


class SubsetFederatedData(FederatedData):
  """A simple wrapper over a concrete FederatedData for restricting to a subset of client ids.

  This is useful when we wish to create a smaller FederatedData out of arbitrary
  client ids, where slicing is not possible.
  """

  def __init__(self,
               base: FederatedData,
               client_ids: Iterable[ClientId],
               validate=True):
    """Initializes the subset federated dataset.

    Args:
      base: Base concrete FederatedData.
      client_ids: Client ids to include in the subset. All client ids must be in
        base.client_ids(), otherwise the behavior of SubsetFederatedData is
        undefined when validate=False.
      validate: Whether to validate client ids.
    """
    self._base = base
    if not isinstance(client_ids, set):
      client_ids = set(client_ids)
    if validate:
      bad_client_ids = client_ids.difference(base.client_ids())
      if bad_client_ids:
        raise ValueError('Some client ids are not in the base FederatedData, '
                         f'showing up to 10: {sorted(bad_client_ids)[:10]}')
    self._client_ids = client_ids

  def slice(self,
            start: Optional[ClientId] = None,
            stop: Optional[ClientId] = None) -> FederatedData:
    if start is None and stop is None:
      client_ids = self._client_ids
    elif start is None:
      client_ids = set(i for i in self._client_ids if i < stop)
    elif stop is None:
      client_ids = set(i for i in self._client_ids if i >= start)
    else:
      client_ids = set(i for i in self._client_ids if start <= i and i < stop)
    return SubsetFederatedData(
        self._base.slice(start, stop), client_ids, validate=False)

  def num_clients(self) -> int:
    return len(self._client_ids)

  def client_ids(self) -> Iterator[ClientId]:
    # Ids are sorted for deterministic iteration order.
    return iter(sorted(self._client_ids))

  def client_sizes(self) -> Iterator[Tuple[ClientId, int]]:
    for client_id, size in self._base.client_sizes():
      if client_id in self._client_ids:
        yield client_id, size

  def client_size(self, client_id: ClientId) -> int:
    if client_id not in self._client_ids:
      raise KeyError
    return self._base.client_size(client_id)

  def clients(self) -> Iterator[Tuple[ClientId, client_datasets.ClientDataset]]:
    # Ids are sorted for deterministic iteration order.
    yield from self.get_clients(sorted(self._client_ids))

  def shuffled_clients(
      self,
      buffer_size: int,
      seed: Optional[int] = None
  ) -> Iterator[Tuple[ClientId, client_datasets.ClientDataset]]:
    rng = np.random.RandomState(seed)
    while True:
      for client_id, dataset in client_datasets.buffered_shuffle(
          self.clients(), buffer_size, rng):
        yield client_id, dataset

  def get_clients(
      self, client_ids: Iterable[ClientId]
  ) -> Iterator[Tuple[ClientId, client_datasets.ClientDataset]]:
    for client_id, dataset in self._base.get_clients(client_ids):
      if client_id not in self._client_ids:
        raise KeyError
      yield client_id, dataset

  def get_client(self, client_id: ClientId) -> client_datasets.ClientDataset:
    if client_id not in self._client_ids:
      raise KeyError
    return self._base.get_client(client_id)

  def preprocess_client(
      self, fn: Callable[[ClientId, client_datasets.Examples],
                         client_datasets.Examples]
  ) -> FederatedData:
    return SubsetFederatedData(
        self._base.preprocess_client(fn), self._client_ids, validate=False)

  def preprocess_batch(
      self, fn: Callable[[client_datasets.Examples], client_datasets.Examples]
  ) -> FederatedData:
    return SubsetFederatedData(
        self._base.preprocess_batch(fn), self._client_ids, validate=False)


class FederatedDataBuilder(abc.ABC):
  """FederatedDataBuilder interface.

  To be implemented as a context manager for building file formats from pairs of
  client IDs and client NumPy examples.

  It is relevant to note that the add method below does not specify any raised
  exceptions. One could imagine some formats where add can fail in some way:
  out-of-order or duplicate inputs, remote files and network failures,
  individual entries too big for a format, etc. In order to address this we let
  implementations throw whatever they see relevant and fit to their particular
  use cases. The same is relevant when it comes to the __init__, __enter__, and
  __exit__ methods, where implementations are left with the responsibility of
  raising exceptions as they see fit to their particular use cases. For example,
  if an invalid file path is passed, or there were any issues finalizing the
  builder, etc.

  Eg of end behavior::

    with FederatedDataBuilder(path) as builder:
      builder.add(b'k1', np.array([b'v1'], dtype=np.object))
      builder.add(b'k2', np.array([b'v2'], dtype=np.object))
  """

  @abc.abstractmethod
  def __enter__(self):
    """Assigns the variable defined after 'as' in the with statement to self.

    By returning self the required functionality is kept within the same class
    so that one can call the add method defined below inside the with block.

    Returns:
      self
    """

  @abc.abstractmethod
  def __exit__(self, exc_type, exc_value, exc_traceback):
    """Finalizes the builder once it leaves the 'with' block.

    Args:
      exc_type: indicates class of exception.
      exc_value: indicates type of exception.
      exc_traceback: traceback is a report which has all of the information
        needed to solve the exception.
    """

  @abc.abstractmethod
  def add(self, client_id: bytes, examples: client_datasets.Examples):
    """Adds a single client ID and client NumPy examples pair to file format.

    Implementations may choose to raise exceptions as appropriate (e.g.,
    out-of-order or duplicate inputs, remote files and network failures,
    individual entries too big for a format, etc.)

    Args:
      client_id: client ID that is associated with the examples parameter.
      examples: single np.ndarray object, or a dict[str, np.ndarray].
    """



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.for_each_client."""

import os

from absl.testing import absltest

from fedjax.core import for_each_client

import jax
from jax.lib import xla_bridge
import jax.numpy as jnp
import numpy as np
import numpy.testing as npt


def setUpModule():
  """Run all tests with 8 CPU devices."""
  global prev_xla_flags  # pylint: disable=global-variable-undefined
  prev_xla_flags = os.getenv('XLA_FLAGS')
  flags_str = prev_xla_flags or ''
  # Don't override user-specified device count, or other XLA flags.
  if 'xla_force_host_platform_device_count' not in flags_str:
    os.environ['XLA_FLAGS'] = (
        flags_str + ' --xla_force_host_platform_device_count=8')
  # Clear any cached backends so new CPU backend will pick up the env var.
  xla_bridge.get_backend.cache_clear()


def tearDownModule():
  """Reset to previous configuration in case other test modules will be run."""
  if prev_xla_flags is None:
    del os.environ['XLA_FLAGS']
  else:
    os.environ['XLA_FLAGS'] = prev_xla_flags
  xla_bridge.get_backend.cache_clear()


# Map over clients and count how many points are greater than `limit` for
# each client. Each client also has a different `start` that is specified via
# client input.


def client_init(shared_input, client_input):
  client_step_state = {
      'limit': shared_input['limit'],
      'count': client_input['start']
  }
  return client_step_state


def client_step(client_step_state, batch):
  num = jnp.sum(batch['x'] > client_step_state['limit'])
  client_step_state = {
      'limit': client_step_state['limit'],
      'count': client_step_state['count'] + num
  }
  return client_step_state


def client_final(shared_input, client_step_state):
  del shared_input  # Unused.
  return client_step_state['count']


# We'll also keep track of the `num` per step in our step results.


def client_step_with_result(client_step_state, batch):
  num = jnp.sum(batch['x'] > client_step_state['limit'])
  client_step_state = {
      'limit': client_step_state['limit'],
      'count': client_step_state['count'] + num
  }
  client_step_result = {'num': num}
  return client_step_state, client_step_result


class DoNotRun:

  class BaseTest(absltest.TestCase):

    @classmethod
    def setUpClass(cls):
      super().setUpClass()
      cls.SHARED_INPUT = {'limit': jnp.array(2)}
      # Three clients with different data and different starting counts.
      cls.CLIENTS = [
          (b'cid0', [{
              'x': jnp.array([1, 2, 3, 4])
          }, {
              'x': jnp.array([1, 2, 3])
          }], {
              'start': jnp.array(2)
          }),
          (b'cid1', [{
              'x': jnp.array([1, 2])
          }, {
              'x': jnp.array([1, 2, 3, 4, 5])
          }], {
              'start': jnp.array(0)
          }),
          (b'cid2', [{
              'x': jnp.array([1])
          }], {
              'start': jnp.array(1)
          }),
      ]
      cls.EXPECTED_WITHOUT_STEP_RESULT = [(b'cid0', jnp.array(5)),
                                          (b'cid1', jnp.array(3)),
                                          (b'cid2', jnp.array(1))]
      cls.EXPECTED_WITH_STEP_RESULT = [
          (b'cid0', jnp.array(5), [{
              'num': jnp.array(2)
          }, {
              'num': jnp.array(1)
          }]),
          (b'cid1', jnp.array(3), [{
              'num': jnp.array(0)
          }, {
              'num': jnp.array(3)
          }]),
          (b'cid2', jnp.array(1), [{
              'num': jnp.array(0)
          }]),
      ]


class ForEachClientTest(DoNotRun.BaseTest):

  def test_without_step_result(self):
    func = for_each_client.for_each_client(client_init, client_step,
                                           client_final)
    results = list(func(self.SHARED_INPUT, self.CLIENTS))
    npt.assert_equal(results, self.EXPECTED_WITHOUT_STEP_RESULT)

  def test_with_step_result(self):
    func = for_each_client.for_each_client(
        client_init,
        client_step_with_result,
        client_final,
        with_step_result=True)
    results = list(func(self.SHARED_INPUT, self.CLIENTS))
    npt.assert_equal(results, self.EXPECTED_WITH_STEP_RESULT)


class ForEachClientJitTest(DoNotRun.BaseTest):

  def setUp(self):
    super().setUp()
    self._backend = for_each_client.ForEachClientJitBackend()

  def test_basic_output(self):
    func = self._backend(client_init, client_step_with_result, client_final)
    results = list(func(self.SHARED_INPUT, self.CLIENTS))
    npt.assert_equal(results, self.EXPECTED_WITH_STEP_RESULT)

  def test_has_jit(self):
    num_calls = [0, 0, 0]

    def my_client_init(*args, **kwargs):
      num_calls[0] += 1
      return client_init(*args, **kwargs)

    def my_client_step(*args, **kwargs):
      num_calls[1] += 1
      return client_step_with_result(*args, **kwargs)

    def my_client_final(*args, **kwargs):
      num_calls[2] += 1
      return client_final(*args, **kwargs)

    func = self._backend(my_client_init, my_client_step, my_client_final)
    npt.assert_equal(
        list(func(self.SHARED_INPUT, self.CLIENTS)),
        self.EXPECTED_WITH_STEP_RESULT)
    self.assertListEqual(num_calls, [1, 5, 1])
    # Has jit, so repeated calls will not increase num_calls.
    npt.assert_equal(
        list(func(self.SHARED_INPUT, self.CLIENTS)),
        self.EXPECTED_WITH_STEP_RESULT)
    self.assertListEqual(num_calls, [1, 5, 1])


class ForEachClientDebugTest(DoNotRun.BaseTest):

  def setUp(self):
    super().setUp()
    self._backend = for_each_client.ForEachClientDebugBackend()

  def test_basic_output(self):
    func = self._backend(client_init, client_step_with_result, client_final)
    results = list(func(self.SHARED_INPUT, self.CLIENTS))
    npt.assert_equal(results, self.EXPECTED_WITH_STEP_RESULT)

  def test_no_jit(self):
    num_calls = [0, 0, 0]

    def my_client_init(*args, **kwargs):
      num_calls[0] += 1
      return client_init(*args, **kwargs)

    def my_client_step(*args, **kwargs):
      num_calls[1] += 1
      return client_step_with_result(*args, **kwargs)

    def my_client_final(*args, **kwargs):
      num_calls[2] += 1
      return client_final(*args, **kwargs)

    func = self._backend(my_client_init, my_client_step, my_client_final)
    npt.assert_equal(
        list(func(self.SHARED_INPUT, self.CLIENTS)),
        self.EXPECTED_WITH_STEP_RESULT)
    self.assertListEqual(num_calls, [3, 5, 3])
    # No jit, so repeated calls will increase num_calls.
    npt.assert_equal(
        list(func(self.SHARED_INPUT, self.CLIENTS)),
        self.EXPECTED_WITH_STEP_RESULT)
    self.assertListEqual(num_calls, [6, 10, 6])

  def test_client_init_error(self):

    def my_client_init(shared_input, client_input):
      if client_input['start'].copy() == 0:
        raise ValueError('Oops')
      return client_init(shared_input, client_input)

    func = self._backend(my_client_init, client_step_with_result, client_final)
    with self.assertRaisesRegex(
        for_each_client.ForEachClientError,
        'Stage: client_init.*Base error is ValueError: Oops') as cm:
      list(func(self.SHARED_INPUT, self.CLIENTS))
    # At least one side of the comparison of npt.assert_equal needs to be
    # np.ndarray to trigger npt.assert_array_equal, thus the device_get calls.
    npt.assert_equal(
        cm.exception.context, {
            'client_id': b'cid1',
            'client_init': my_client_init,
            'shared_input': jax.device_get(self.SHARED_INPUT),
            'client_input': jax.device_get({'start': jnp.array(0)})
        })

  def test_client_step_error(self):

    def my_client_step(state, batch):
      if len(batch['x']) == 3:
        raise ValueError('Oops')
      return client_step_with_result(state, batch)

    func = self._backend(client_init, my_client_step, client_final)
    with self.assertRaisesRegex(
        for_each_client.ForEachClientError,
        r'Stage: client_step.*Base error is ValueError: Oops') as cm:
      list(func(self.SHARED_INPUT, self.CLIENTS))
    # At least one side of the comparison of npt.assert_equal needs to be
    # np.ndarray to trigger npt.assert_array_equal, thus the device_get calls.
    npt.assert_equal(
        cm.exception.context, {
            'client_id':
                b'cid0',
            'client_step':
                my_client_step,
            'state':
                jax.device_get({
                    'limit': jnp.array(2),
                    'count': jnp.array(4)
                }),
            'batch':
                jax.device_get({'x': jnp.array([1, 2, 3])})
        })

  def test_client_final_error(self):

    def my_client_final(shared_input, state):
      if state['count'].copy() == 1:
        raise ValueError('Oops')
      return client_final(shared_input, state)

    func = self._backend(client_init, client_step_with_result, my_client_final)
    with self.assertRaisesRegex(
        for_each_client.ForEachClientError,
        r'Stage: client_final.*Base error is ValueError: Oops') as cm:
      list(func(self.SHARED_INPUT, self.CLIENTS))
    # At least one side of the comparison of npt.assert_equal needs to be
    # np.ndarray to trigger npt.assert_array_equal, thus the device_get calls.
    npt.assert_equal(
        cm.exception.context, {
            'client_id':
                b'cid2',
            'client_final':
                my_client_final,
            'shared_input':
                jax.device_get(self.SHARED_INPUT),
            'state':
                jax.device_get({
                    'limit': jnp.array(2),
                    'count': jnp.array(1)
                })
        })


class BlockifyTest(absltest.TestCase):

  def test_blockify(self):
    clients = [
        ('a', np.random.uniform(size=(1, 8)), np.random.uniform(size=(2,))),
        ('b', np.random.uniform(size=(4, 8)), np.random.uniform(size=(2,))),
        ('c', np.random.uniform(size=(3, 8)), np.random.uniform(size=(2,))),
        ('d', np.random.uniform(size=(2, 8)), np.random.uniform(size=(2,)))
    ]
    a, b, c, d = [clients[i][1] for i in range(4)]

    with self.subTest('no padding client'):
      blocks = list(for_each_client._blockify(clients, 2))
      self.assertLen(blocks, 2)
      # block 0
      self.assertListEqual(blocks[0].client_id, ['b', 'c'])
      self.assertListEqual(blocks[0].client_mask, [True, True])
      self.assertListEqual(blocks[0].num_batches, [4, 3])
      npt.assert_equal(blocks[0].masked_batches,
                       [([b[0], c[0]], [True, True]),
                        ([b[1], c[1]], [True, True]),
                        ([b[2], c[2]], [True, True]),
                        ([b[3], np.zeros_like(b[0])], [True, False])])
      npt.assert_equal(blocks[0].client_input, [clients[1][-1], clients[2][-1]])
      # block 1
      self.assertListEqual(blocks[1].client_id, ['d', 'a'])
      self.assertListEqual(blocks[1].client_mask, [True, True])
      self.assertListEqual(blocks[1].num_batches, [2, 1])
      npt.assert_equal(blocks[1].masked_batches,
                       [([d[0], a[0]], [True, True]),
                        ([d[1], np.zeros_like(d[0])], [True, False])])
      npt.assert_equal(blocks[1].client_input, [clients[3][-1], clients[0][-1]])

    with self.subTest('has padding client'):
      blocks = list(for_each_client._blockify(clients, 3))
      self.assertLen(blocks, 2)
      # block 0
      self.assertListEqual(blocks[0].client_id, ['b', 'c', 'd'])
      self.assertListEqual(blocks[0].client_mask, [True, True, True])
      self.assertListEqual(blocks[0].num_batches, [4, 3, 2])
      npt.assert_equal(
          blocks[0].masked_batches,
          [([b[0], c[0], d[0]], [True, True, True]),
           ([b[1], c[1], d[1]], [True, True, True]),
           ([b[2], c[2], np.zeros_like(b[0])], [True, True, False]),
           ([b[3], np.zeros_like(b[0]),
             np.zeros_like(b[0])], [True, False, False])])
      npt.assert_equal(blocks[0].client_input,
                       [clients[1][-1], clients[2][-1], clients[3][-1]])
      # block 1
      self.assertListEqual(blocks[1].client_id, ['a', None, None])
      self.assertListEqual(blocks[1].client_mask, [True, False, False])
      self.assertListEqual(blocks[1].num_batches, [1, 0, 0])
      npt.assert_equal(blocks[1].masked_batches,
                       [([a[0], np.zeros_like(a[0]),
                          np.zeros_like(a[0])], [True, False, False])])
      npt.assert_equal(blocks[1].client_input, [
          clients[0][-1],
          np.zeros_like(clients[0][-1]),
          np.zeros_like(clients[0][-1])
      ])

  def test_blockify_zero_batches(self):
    blocks = list(for_each_client._blockify([('a', [], np.array(1))], 3))
    self.assertLen(blocks, 1)
    self.assertListEqual(blocks[0].client_id, ['a', None, None])
    self.assertListEqual(blocks[0].client_mask, [True, False, False])
    self.assertListEqual(blocks[0].num_batches, [0, 0, 0])
    self.assertListEqual(blocks[0].masked_batches, [])
    npt.assert_equal(
        blocks[0].client_input,
        [np.array(1), np.array(0), np.array(0)])


class ForEachClientPmapTest(absltest.TestCase):

  def test_for_each_client_pmap(self):
    # Make sure setUpModule() does the work.
    self.assertEqual(jax.local_device_count(), 8)

    def my_client_init(shared_input, client_input):
      return {'x': jnp.dot(shared_input['y'], client_input['z'])}

    def my_client_step(state, batch):
      return {'x': jnp.dot(state['x'], batch['w'])}, {'u': jnp.sum(batch['w'])}

    def my_client_final(shared_input, state):
      return {'v': jnp.dot(state['x'], shared_input['y'])}

    shared_input = {'y': np.random.uniform(size=[16, 16])}
    clients = []
    for i in range(10):
      client_id = i
      client_input = {'z': np.random.uniform(size=[16, 16])}
      client_batches = []
      for _ in range(i):
        client_batches.append({'w': np.random.uniform(size=[16, 16])})
      clients.append((client_id, client_batches, client_input))

    expected = {}
    for client_id, client_output, step_results in (
        for_each_client.ForEachClientJitBackend()(my_client_init,
                                                  my_client_step,
                                                  my_client_final)(shared_input,
                                                                   clients)):
      expected[client_id] = jax.device_get((client_output, step_results))

    for i in range(jax.local_device_count() + 1):
      with self.subTest(f'{i} devices' if i > 0 else 'default devices'):
        if i > 0:
          devices = jax.local_devices()[:i]
        else:
          devices = None
        actual = {}
        for client_id, client_output, step_results in (
            for_each_client.ForEachClientPmapBackend(devices)(
                my_client_init, my_client_step, my_client_final)(shared_input,
                                                                 clients)):
          actual[client_id] = (client_output, step_results)
        jax.tree_util.tree_multimap(npt.assert_allclose, actual, expected)
        # Check actual can be operated over.
        jax.tree_util.tree_multimap(
            npt.assert_allclose,
            *jax.tree_util.tree_map(lambda x: x + 1, (actual, expected)))


class BackendChoiceTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self._backend = for_each_client.ForEachClientJitBackend()

  def tearDown(self):
    super().tearDown()
    for_each_client.set_for_each_client_backend(None)

  def test_default_backend(self):
    self.assertIsInstance(for_each_client.get_for_each_client_backend(),
                          for_each_client.ForEachClientJitBackend)

  def test_set_and_get_concrete(self):
    self.assertIsNot(for_each_client.get_for_each_client_backend(),
                     self._backend)
    for_each_client.set_for_each_client_backend(self._backend)
    self.assertIs(for_each_client.get_for_each_client_backend(), self._backend)

  def test_set_and_get_str(self):
    with self.subTest('debug'):
      for_each_client.set_for_each_client_backend('debug')
      self.assertIsInstance(for_each_client.get_for_each_client_backend(),
                            for_each_client.ForEachClientDebugBackend)
    with self.subTest('jit'):
      for_each_client.set_for_each_client_backend('jit')
      self.assertIsInstance(for_each_client.get_for_each_client_backend(),
                            for_each_client.ForEachClientJitBackend)
    with self.subTest('pmap'):
      for_each_client.set_for_each_client_backend('pmap')
      self.assertIsInstance(for_each_client.get_for_each_client_backend(),
                            for_each_client.ForEachClientPmapBackend)
    with self.subTest('invalid'):
      with self.assertRaisesRegex(ValueError, "Unsupported backend 'invalid'"):
        for_each_client.set_for_each_client_backend('invalid')

  def test_context_manager(self):
    with for_each_client.for_each_client_backend(self._backend):
      self.assertIs(for_each_client.get_for_each_client_backend(),
                    self._backend)
      with for_each_client.for_each_client_backend('debug'):
        self.assertIsInstance(for_each_client.get_for_each_client_backend(),
                              for_each_client.ForEachClientDebugBackend)
      self.assertIs(for_each_client.get_for_each_client_backend(),
                    self._backend)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for evaluation."""

import collections
from typing import Dict, Iterable, Iterator, List, Optional

from fedjax.legacy.core import dataset_util
from fedjax.legacy.core import metrics
from fedjax.legacy.core import prefetch
from fedjax.legacy.core.model import Model
from fedjax.legacy.core.typing import FederatedData
from fedjax.legacy.core.typing import MetricResults
from fedjax.legacy.core.typing import Params


def aggregate_metrics(
    metrics_iter: Iterable[Dict[str, metrics.Metric]]) -> MetricResults:
  """Aggregates multiple metrics into a single overall metrics output.

  Args:
    metrics_iter: Multiple mappings of metrics.

  Returns:
    Aggregated result metrics or empty mapping if `metrics_iter` is empty.
  """
  metrics_iter = iter(metrics_iter)
  try:
    aggregated_metrics = next(metrics_iter)
  except StopIteration:
    return collections.OrderedDict()

  for metrics_dict in metrics_iter:
    next_aggregated_metrics = collections.OrderedDict()
    for name, metric in metrics_dict.items():
      next_aggregated_metrics[name] = aggregated_metrics[name].merge(metric)
    aggregated_metrics = next_aggregated_metrics

  metrics_results = collections.OrderedDict()
  for name, metric in aggregated_metrics.items():
    metrics_results[name] = metric.result()

  return metrics_results


def evaluate_single_client(dataset: dataset_util.DatasetOrIterable,
                           model: Model, params: Params) -> MetricResults:
  """Evaluates model for a single client's dataset.

  Args:
    dataset: Dataset of ordered mapping of examples [client_batch_size, d].
    model: Model implementation.
    params: Model parameters.

  Returns:
    Ordered mapping of client metrics or empty mapping if the dataset is empty.
  """
  def compute_batch_metrics(batch):
    return model.evaluate(params, batch)
  batch_metrics_iterator = map(compute_batch_metrics,
                               dataset_util.iterate(dataset))
  return aggregate_metrics(batch_metrics_iterator)


def evaluate_multiple_clients(
    federated_data: FederatedData,
    client_ids: List[str],
    model: Model,
    params: Params,
    client_data_hparams: Optional[dataset_util.ClientDataHParams] = None
) -> Iterator[MetricResults]:
  """Evaluates model over input clients' datasets.

  Args:
    federated_data: Federated data separated per client.
    client_ids: Ids of clients to evaluate.
    model: Model implementation.
    params: Pytree of model parameters to be evaluated.
    client_data_hparams: Hyperparameters for client dataset preparation.

  Yields:
    Ordered mapping of metrics per client or empty mapping if a client's dataset
      is emtpy.
  """
  if client_data_hparams is not None:
    federated_data = federated_data.preprocess(
        lambda ds: dataset_util.preprocess_tf_dataset(ds, client_data_hparams))
  for _, client_dataset in prefetch.PrefetchClientDatasetsIterator(
      federated_data, client_ids):
    yield evaluate_single_client(client_dataset, model, params)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A small library for working with evaluation metrics such as accuracy."""

import abc
import functools
from typing import Optional, Tuple

from fedjax.core import dataclasses
from fedjax.core import util
from fedjax.core.typing import BatchExample
from fedjax.core.typing import BatchPrediction
from fedjax.core.typing import SingleExample
from fedjax.core.typing import SinglePrediction

import jax
import jax.numpy as jnp


class Stat(metaclass=abc.ABCMeta):
  """Stat keeps some statistic, along with operations over them.

  Most users will only need to interact with a :class:`Stat` object via
  :py:meth:`~Stat.result`

  For those who need to create new metrics, please first read the
  :ref:`under-the-hood` section of the module docstring.

  Most :class:`Stat`'s domain (the set of possible statistic values) has
  constraints, it is thus usually a good practice to offer and use factory
  methods to construct new :class:`Stat` objects instead of directly assigning
  the fields.

  To work with various jax constructs, a concrete :class:`Stat` should be a
  PyTree.
  This is easily achieved with ``fedjax.dataclass``.

  A :class:`Stat` may hold either a single statistic (a rank 0 :class:`Stat`),
  or an array of statistics (a higher rank :class:`Stat`).
  :py:meth:`~Stat.result` and :py:meth:`~Stat.merge` only needs to
  work on a rank 0 :class:`Stat` :py:meth:`~Stat.reduce` only needs to work on
  a higher rank :class:`Stat`
  """

  @abc.abstractmethod
  def result(self) -> jnp.ndarray:
    """Calculates the metric value from the statistic value.

    For example, :meth:`MeanStat.result` calculates a weighted average.

    Returns:
      The return value must be a ``jnp.ndarray``.
    """

  @abc.abstractmethod
  def merge(self, other: 'Stat') -> 'Stat':
    """Merges two Stat objects into a new Stat with merged statistics.

    Args:
      other: Another Stat object of the same type.

    Returns:
      A new Stat object of the same type with merged statistics.
    """

  @abc.abstractmethod
  def reduce(self, axis: Optional[int] = 0) -> 'Stat':
    """Reduces a higher rank statistic along a given ``axis``.

    See the class docstring for details.

    Args:
      axis: An integer axis index, or ``None``.

    Returns:
      A new Stat object of the same type.
    """

  def __str__(self) -> str:
    return f'{repr(self)} => {self.result()}'


@dataclasses.dataclass
class MeanStat(Stat):
  """Statistic for weighted mean calculation.

  Prefer using the :meth:`MeanStat.new()` factory method instead of directly
  assigning to fields.

  Example::

    stat_0 = MeanStat.new(accum=1, weight=2)
    stat_1 = MeanStat.new(accum=2, weight=3)
    merged_stat = stat_0.merge(stat_1)
    print(merged_stat)
    # MeanState(accum=3, weight=5) => 0.6

    stat = MeanStat.new(jnp.array([1, 2, 4]), jnp.array([1, 1, 0]))
    reduced_stat = stat.reduce()
    print(reduced_stat)
    # MeanStat(accum=3, weight=2) => 1.5

  Attributes:
    accum: The weighted sum.
    weight: The sum of weights.
  """
  accum: jnp.ndarray
  weight: jnp.ndarray

  @classmethod
  def new(cls, accum, weight) -> 'MeanStat':
    """Creates a sanitized MeanStat.

    The domain of a weighted mean statistic is:

    .. math::

      \{(0, 0)\} ∪ \{(a, b) | a >= 0, b > 0\}


    new() sanitizes values outside the domain into the identity (zeros).

    Args:
      accum: A value convertible to ``jnp.ndarray``.
      weight: A value convertible to ``jnp.ndarray``.

    Returns:
      The sanitized MeanStat.
    """
    weight = jnp.maximum(0, jnp.array(weight, copy=False))
    accum = jnp.where(weight == 0, 0, jnp.array(accum, copy=False))
    return cls(accum, weight)

  def result(self) -> jnp.ndarray:
    return util.safe_div(self.accum, self.weight)

  def merge(self, other: 'MeanStat') -> 'MeanStat':
    accum = self.accum + other.accum
    weight = self.weight + other.weight
    return MeanStat.new(accum, weight)

  def reduce(self, axis: Optional[int] = 0) -> 'MeanStat':
    return MeanStat.new(
        jnp.sum(self.accum, axis=axis), jnp.sum(self.weight, axis=axis))


@dataclasses.dataclass
class SumStat(Stat):
  """Statistic for summing values.

  Example::

    stat_0 = SumStat.new(accum=1)
    stat_1 = SumStat.new(accum=2)
    merged_stat = stat_0.merge(stat_1)
    print(merged_stat)
    # SumStat(accum=3) => 3

    stat = SumStat.new(jnp.array([1, 2, 1]))
    reduced_stat = stat.reduce()
    print(reduced_stat)
    # SumStat(accum=4) => 4

  Attributes:
    accum: Sum of values.
  """
  accum: jnp.ndarray

  @classmethod
  def new(cls, accum: jnp.ndarray) -> 'SumStat':
    """Creates a sanitized SumStat."""
    return cls(jnp.array(accum, copy=False))

  def result(self) -> jnp.ndarray:
    return self.accum

  def merge(self, other: 'SumStat') -> 'SumStat':
    return SumStat.new(self.accum + other.accum)

  def reduce(self, axis: Optional[int] = 0) -> 'SumStat':
    return SumStat.new(jnp.sum(self.accum, axis=axis))


class Metric(metaclass=abc.ABCMeta):
  """Metric is the conceptual metric (like accuracy).

  It defines two methods:

  - :meth:`~Metric.evaluate_example` evaluates a single example, and returns a
    :class:`Stat` object.
  - :meth:`~Metric.zero` returns the identity value for what
    :meth:`~Metric.evaluate_example` returns.

  Given a :class:`Metric` object ``m``, let

  - ``u = m.zero()``
  - ``v = m.evaluate_example(...)``

  We require that

  - ``type(u) == type(v)``.
  - ``u.merge(v) == v.merge(u) == v``.
  - Components of ``u`` has the same shape as the counter parts in ``v``.
  """

  @abc.abstractmethod
  def zero(self) -> Stat:
    """Returns a Stat such that merging with it is an identity operation.

    e.g. for accuracy: ``MeanStat.new(0., 0.)``

    Returns:
      Stat identity value.
    """

  @abc.abstractmethod
  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> Stat:
    """Evaluates a single example.

    e.g. for accuracy: ``MeanStat.new(num_correct, num_total)``

    Args:
      example: A single input example (e.g. one sentence for language).
      prediction: Output for ``example`` from
        :meth:`fedjax.core.models.Model.apply_for_eval`.

    Returns:
      Stat value.
    """


def apply_mask(mask: jnp.ndarray, a: jnp.ndarray,
               b: jnp.ndarray) -> jnp.ndarray:
  """Applies mask on the leading dimension."""
  rank = max(len(a.shape), len(b.shape))
  return jnp.where(jnp.expand_dims(mask, tuple(range(1, rank))), a, b)


@functools.partial(jax.jit, static_argnums=0)
def evaluate_batch(metric: Metric,
                   batch_example: BatchExample,
                   batch_prediction: BatchPrediction,
                   batch_mask: Optional[jnp.ndarray] = None) -> Stat:
  """Evaluates a batch using a metric."""
  batch_stat = jax.vmap(metric.evaluate_example)(batch_example,
                                                 batch_prediction)
  if batch_mask is not None:
    batch_stat = jax.tree_util.tree_multimap(
        functools.partial(apply_mask, batch_mask), batch_stat, metric.zero())
  return batch_stat.reduce()


def unreduced_cross_entropy_loss(targets: jnp.ndarray,
                                 preds: jnp.ndarray) -> jnp.ndarray:
  """Returns unreduced cross entropy loss."""
  num_classes = preds.shape[-1]
  log_preds = jax.nn.log_softmax(preds)
  one_hot_targets = jax.nn.one_hot(targets, num_classes)
  return -jnp.sum(one_hot_targets * log_preds, axis=-1)


@dataclasses.dataclass
class CrossEntropyLoss(Metric):
  """Metric for cross entropy loss.

  Example::

    example = {'y': jnp.array(1)}
    prediction = jnp.array([1.2, 0.4])
    metric = CrossEntropyLoss()
    print(metric.evaluate_example(example, prediction))
    # MeanStat(accum=1.1711007, weight=1) => 1.1711007

  Attributes:
    target_key: Key name in ``example`` for target.
    pred_key: Key name in ``prediction`` for unnormalized model output pred.
  """
  target_key: str = 'y'
  pred_key: Optional[str] = None

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes cross entropy loss for a single example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for ``example`` of shape [num_classes]

    Returns:
      MeanStat for loss for a single example.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    loss = unreduced_cross_entropy_loss(target, pred)
    return MeanStat.new(loss, 1.)


@dataclasses.dataclass
class Accuracy(Metric):
  """Metric for accuracy.

  Example::

    example = {'y': jnp.array(2)}
    prediction = jnp.array([0, 0, 1])
    metric = Accuracy()
    print(metric.evaluate_example(example, prediction))
    # MeanStat(accum=1, weight=1) => 1

  Attributes:
    target_key: Key name in ``example`` for target.
    pred_key: Key name in ``prediction`` for unnormalized model output pred.
  """
  target_key: str = 'y'
  pred_key: Optional[str] = None

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes accuracy for a single example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for ``example`` of shape [num_classes]

    Returns:
      MeanStat for accuracy for a single example.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    correct = (target == jnp.argmax(pred, axis=-1)).astype(jnp.float32)
    return MeanStat.new(correct, 1.)


def get_target_weight(target: jnp.ndarray,
                      masked_target_values: Tuple[int, ...]) -> jnp.ndarray:
  target_weight = jnp.ones_like(target, dtype=jnp.float32)
  for mv in masked_target_values:
    target_weight *= (target != mv)
  return target_weight


@dataclasses.dataclass
class TopKAccuracy(Metric):
  """Metric for top k accuracy.

  This metric computes the number of times where the correct class
  is among the top k classes predicted.

  Example: top 3 accuracy

  - Dog => [Dog, Cat, Bird, Mouse, Penguin] ✓
  - Cat => [Bird, Mouse, Cat, Penguin, Dog] ✓
  - Dog => [Dog, Cat, Bird, Penguin, Mouse] ✓
  - Bird => [Bird, Cat, Mouse, Penguin, Dog] ✓
  - Cat => [Cat, Bird, Mouse, Dog, Penguin] ✓
  - Cat => [Cat, Mouse, Dog, Penguin, Bird] ✓
  - Mouse => [Penguin, Cat, Dog, Mouse, Bird] x
  - Penguin => [Dog, Mouse, Cat, Penguin, Bird] x

  6 correct predictions in top 3 predicted classes / 8 total examples
  = .75 top 3 accuracy

  Top k accuracy, also known as top n accuracy,
  is a useful metric when it comes to recommendations.
  One example would be the word recommendations on a virtual keyboard
  where three suggested words are displayed.

  For k=1, we strongly recommend using :class:`Accuracy` to avoid an
  unnecessary argsort. k < 1 will return 0. and k >= num_classes will
  return 1.

  If two or more classes have the same prediction, the classes will be
  considered in order of lowest to highest indices.

  Example::

    example = {'y': jnp.array(2)}
    prediction = jnp.array([0, 0.5, 0.2])
    metric = TopKAccuracy(k=2)
    print(metric.evaluate_example(example, prediction))
    # MeanStat(accum=1, weight=1) => 1

  Attributes:
    k: Number of top elements to look at for computing accuracy.
    target_key: Key name in ``example`` for target.
    pred_key: Key name in ``prediction`` for unnormalized model output pred.
  """
  k: int
  target_key: str = 'y'
  pred_key: Optional[str] = None

  def zero(self) -> Stat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> Stat:

    """Computes top k accuracy for a single example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for ``example`` of shape [num_clases].

    Returns:
      MeanStat for top k accuracy for a single example.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    top_k_pred = jnp.argsort(-pred)[:self.k]
    correct = jnp.any(top_k_pred == target).astype(jnp.float32)
    return MeanStat.new(correct, 1.)


@dataclasses.dataclass
class SequenceTokenCrossEntropyLoss(Metric):
  """Metric for token cross entropy loss for a sequence example.

  Example::

    example = {'y': jnp.array([1, 0, 1])}
    prediction = jnp.array([[1.2, 0.4], [2.3, 0.1], [0.3, 3.2]])
    metric = SequenceTokenCrossEntropyLoss()
    print(metric.evaluate_example(example, prediction))
    # MeanStat(accum=1.2246635, weight=2) => 0.61233175

    per_position_metric = SequenceTokenCrossEntropyLoss(per_position=True)
    print(per_position_metric.evaluate_example(example, prediction))
    # MeanStat(accum=[1.1711007, 0., 0.05356275], weight=[1., 0., 1.]) => [1.1711007, 0., 0.05356275]

  Attributes:
    target_key: Key name in ``example`` for target.
    pred_key: Key name in ``prediction`` for unnormalized model output pred.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
    per_position: Whether to keep output statistic per position or sum across
      positions for the entire sequence.
  """
  target_key: str = 'y'
  pred_key: Optional[str] = None
  masked_target_values: Tuple[int, ...] = (0,)
  per_position: bool = False

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes token cross entropy loss for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for ``example`` of shape [num_classes]

    Returns:
      MeanStat for token loss for either a single sequence example or at each
        token position if ``per_position`` is ``True``.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    target_weight = get_target_weight(target, self.masked_target_values)
    token_loss = unreduced_cross_entropy_loss(target, pred)
    if self.per_position:
      return MeanStat.new(token_loss * target_weight, target_weight)
    return MeanStat.new(
        jnp.sum(token_loss * target_weight), jnp.sum(target_weight))


@dataclasses.dataclass
class SequenceCrossEntropyLoss(Metric):
  """Metric for total cross entropy loss for a sequence example.

  Example::

    example = {'y': jnp.array([1, 0, 1])}
    prediction = jnp.array([[1.2, 0.4], [2.3, 0.1], [0.3, 3.2]])
    metric = SequenceCrossEntropyLoss()
    print(metric.evaluate_example(example, prediction))
    # MeanStat(accum=1.2246635, weight=1) => 1.2246635

  Attributes:
    target_key: Key name in ``example`` for target.
    pred_key: Key name in ``prediction`` for unnormalized model output pred.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
  """
  target_key: str = 'y'
  pred_key: Optional[str] = None
  masked_target_values: Tuple[int, ...] = (0,)

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes total cross entropy loss for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for ``example`` of shape [num_classes]

    Returns:
      MeanStat for total loss for a single sequence example.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    target_weight = get_target_weight(target, self.masked_target_values)
    token_loss = unreduced_cross_entropy_loss(target, pred)
    # Change weight from number of non masked target tokens to 1 if the sequence
    # contains any non masked tokens or 0 if the entire sequence is masked.
    return MeanStat.new(
        jnp.sum(token_loss * target_weight), jnp.any(target_weight))


@dataclasses.dataclass
class SequenceTokenAccuracy(Metric):
  """Metric for token accuracy for a sequence example.

  Example::

    example = {'y': jnp.array([1, 2, 2, 1, 3, 0])}
    # prediction = [1, 0, 2, 1, 3, 0].
    prediction = jnp.array([[0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 1, 0],
                            [0, 1, 0, 0], [0, 0, 0, 1], [1, 0, 0, 0]])
    logits_mask = (0., 0., 0., jnp.NINF)
    metric = SequenceTokenAccuracy(logits_mask=logits_mask)
    print(metric.evaluate_example(example, prediction))
    # MeanStat(accum=3, weight=5) => 0.6

    per_position_metric = SequenceTokenAccuracy(logits_mask=logits_mask, per_position=True)
    print(per_position_metric.evaluate_example(example, prediction))
    # MeanStat(accum=[1., 0., 1., 1., 0., 0.], weight=[1., 1., 1., 1., 1., 0.]) => [1., 0., 1., 1., 0., 0.]

  Attributes:
    target_key: Key name in ``example`` for target.
    pred_key: Key name in ``prediction`` for unnormalized model output pred.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
    logits_mask: Mask of shape [num_classes] to be applied for preds. This is
      typically used to discount predictions for out-of-vocabulary tokens.
    per_position: Whether to keep output statistic per position or sum across
      positions for the entire sequence.
  """
  target_key: str = 'y'
  pred_key: Optional[str] = None
  masked_target_values: Tuple[int, ...] = (0,)
  # logits_mask cannot be a jnp.ndarray nor np.ndarray because they are not
  # hashable and `Metric`s must be hashable for `evaluate_model`.
  logits_mask: Optional[Tuple[float, ...]] = None
  per_position: bool = False

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes token accuracy for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for ``example`` of shape [num_classes]

    Returns:
      MeanStat for token accuracy for a single sequence example or at each
        token position if ``per_position`` is ``True``.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    if self.logits_mask is not None:
      logits_mask = jnp.array(self.logits_mask)
      pred += logits_mask
    target_weight = get_target_weight(target, self.masked_target_values)
    correct = (target == jnp.argmax(pred, axis=-1)).astype(jnp.float32)
    if self.per_position:
      return MeanStat.new(correct * target_weight, target_weight)
    return MeanStat.new(
        jnp.sum(correct * target_weight), jnp.sum(target_weight))


@dataclasses.dataclass
class SequenceTokenTopKAccuracy(Metric):
  """Metric for token top k accuracy for a sequence example.

  For more information on the top k accuracy metric,
  refer to the :class:`TopKAccuracy` docstring.

  Example::

    example = {'y': jnp.array([1, 2, 2, 1, 3, 0])}
    prediction = jnp.array([[0, 1, 0.5, 0], [1, 0.5, 0, 0], [0.8, 0, 0.7, 0],
                            [0.5, 1, 0, 0], [0, 0.5, 0, 1], [0.5, 0, 0.9, 0]])
    logits_mask = (0., 0., 0., jnp.NINF)
    metric = SequenceTokenTopKAccuracy(k=2, logits_mask=logits_mask)
    print(metric.evaluate_example(example, prediction))
    # MeanStat(accum=3, weight=5) => 0.6

    per_position_metric = SequenceTokenTopKAccuracy(k=2, logits_mask=logits_mask, per_position=True)
    print(per_position_metric.evaluate_example(example, prediction))
    # MeanStat(accum=[1., 0., 1., 1., 0., 0.], weight=[1., 1., 1., 1., 1., 0.]) => [1., 0., 1., 1., 0., 0.]

  Attributes:
    k: Number of top elements to look at for computing accuracy.
    target_key: Key name in ``example`` for target.
    pred_key: Key name in ``prediction`` for unnormalized model output pred.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
    logits_mask: Mask of shape [num_classes] to be applied for preds. This is
      typically used to discount predictions for out-of-vocabulary tokens.
    per_position: Whether to keep output statistic per position or sum across
      positions for the entire sequence.
  """
  k: int
  target_key: str = 'y'
  pred_key: Optional[str] = None
  masked_target_values: Tuple[int, ...] = (0,)
  # logits_mask cannot be a jnp.ndarray nor np.ndarray because they are not
  # hashable and `Metric`s must be hashable for `evaluate_model`.
  logits_mask: Optional[Tuple[float, ...]] = None
  per_position: bool = False

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes token top k accuracy for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape
        [max_length].
      prediction: Unnormalized prediction for ``example`` of shape
        [max_length, num_classes]

    Returns:
      MeanStat for token top k accuracy for a single sequence example or at each
        token position if ``per_position`` is ``True``.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    if self.logits_mask is not None:
      logits_mask = jnp.array(self.logits_mask)
      pred += logits_mask
    target_weight = get_target_weight(target, self.masked_target_values)
    top_k_pred = jnp.argsort(-pred, axis=1)[:, :self.k]
    correct = jnp.any(
        jnp.transpose(top_k_pred) == target, axis=0).astype(jnp.float32)
    if self.per_position:
      return MeanStat.new(correct * target_weight, target_weight)
    return MeanStat.new(
        jnp.sum(correct * target_weight), jnp.sum(target_weight))


@dataclasses.dataclass
class SequenceTokenCount(Metric):
  """Metric for count of non masked tokens for a sequence example.

  Example::

    example = {'y': jnp.array([1, 2, 2, 3, 4, 0, 0])}
    prediction = jnp.array([])  # Unused.
    metric = SequenceTokenCount(masked_target_values=(0, 2))
    print(metric.evaluate_example(example, prediction))
    # SumStat(accum=3) => 3

  Attributes:
    target_key: Key name in ``example`` for target.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
  """
  target_key: str = 'y'
  masked_target_values: Tuple[int, ...] = (0,)

  def zero(self) -> SumStat:
    return SumStat.new(0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> SumStat:
    """Computes total number of non masked tokens in a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for ``example`` of shape [num_classes]

    Returns:
      SumStat for count of non masked tokens for a single sequence example.
    """
    del prediction
    target = example[self.target_key]
    target_weight = get_target_weight(target, self.masked_target_values)
    return SumStat.new(jnp.sum(target_weight))


@dataclasses.dataclass
class SequenceCount(Metric):
  """Metric for count of non masked sequences.

  Example::

    example = {'y': jnp.array([1, 2, 2, 3, 4, 0, 0])}
    empty_example = {'y': jnp.array([0, 0, 0, 0, 0, 0, 0])}
    prediction = jnp.array([])  # Unused.
    metric = metrics.SequenceCount(masked_target_values=(0, 2))
    print(metric.evaluate_example(example, prediction))
    # SumStat(accum=1)
    print(metric.evaluate_example(empty_example, prediction))
    # SumStat(accum=0)

  Attributes:
    target_key: Key name in ``example`` for target.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
  """
  target_key: str = 'y'
  masked_target_values: Tuple[int, ...] = (0,)

  def zero(self) -> SumStat:
    return SumStat.new(0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> SumStat:
    """Counts non masked sequences.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for ``example`` of shape [num_classes]

    Returns:
      SumStat for count of non masked sequence. This will be 0 or 1.
    """
    del prediction
    target = example[self.target_key]
    target_weight = get_target_weight(target, self.masked_target_values)
    return SumStat.new(jnp.any(target_weight).astype(jnp.float32))


@dataclasses.dataclass
class SequenceTruncationRate(Metric):
  """Metric for truncation rate for a sequence example.

  Example::

    example = {'y': jnp.array([1, 2, 2, 3, 3, 3, 4])}
    truncated_example = {'y': jnp.array([1, 2, 2, 3, 3, 3, 3])}
    prediction = jnp.array([])  # Unused.
    metric = SequenceTruncationRate(eos_target_value=4)
    print(metric.evaluate_example(example, prediction))
    # MeanStat(accum=0, weight=1) => 0
    print(metric.evaluate_example(truncated_example, prediction))
    # MeanStat(accum=1, weight=1) => 1

  Attributes:
    eos_target_value: Target value denoting end of sequence. Truncated sequences
      will not have this value.
    target_key: Key name in ``example`` for target.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
  """
  eos_target_value: int
  target_key: str = 'y'
  masked_target_values: Tuple[int, ...] = (0,)

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes truncation rate for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for ``example`` of shape [num_classes]

    Returns:
      MeanStat for truncation rate for a single sequence.
    """
    del prediction
    target = example[self.target_key]
    target_weight = get_target_weight(target, self.masked_target_values)
    not_empty = jnp.sum(jnp.any(target_weight))
    target_is_truncated = jnp.all(target != self.eos_target_value)
    return MeanStat.new(target_is_truncated * not_empty, not_empty)


@dataclasses.dataclass
class SequenceTokenOOVRate(Metric):
  """Metric for out-of-vocabulary (OOV) rate for a sequence example.

  Example::

    example = {'y': jnp.array([1, 2, 2, 3, 4, 0, 0])}
    prediction = jnp.array([])  # Unused.
    metric = SequenceTokenOOVRate(oov_target_values=(2,))
    print(metric.evaluate_example(example, prediction))
    # MeanStat(accum=2, weight=5) => 0.4

    per_position_metric = SequenceTokenOOVRate(oov_target_values=(2,), per_position=True)
    print(per_position_metric.evaluate_example(example, prediction))
    # MeanStat(accum=[0., 1., 1., 0., 0., 0., 0.], weight=[1., 1., 1., 1., 1., 0., 0.]) => [0. 1. 1. 0. 0. 0. 0.]

  Attributes:
    oov_target_values: Target values denoting out-of-vocabulary values.
    target_key: Key name in ``example`` for target.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
    per_position: Whether to keep output statistic per position or sum across
      positions for the entire sequence.
  """
  oov_target_values: Tuple[int, ...]
  target_key: str = 'y'
  masked_target_values: Tuple[int, ...] = (0,)
  per_position: bool = False

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes token out of vocabulary rate for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for ``example`` of shape [num_classes]

    Returns:
      MeanStat for token out of vocabulary rate for a single sequence or at each
        token position if ``per_position`` is ``True``.
    """
    del prediction
    target = example[self.target_key]
    target_weight = get_target_weight(target, self.masked_target_values)
    target_oov = jnp.ones_like(target, dtype=jnp.float32)
    for oov_value in self.oov_target_values:
      target_oov *= (target == oov_value)
    if self.per_position:
      return MeanStat.new(target_oov * target_weight, target_weight)
    return MeanStat.new(
        jnp.sum(target_oov * target_weight), jnp.sum(target_weight))


@dataclasses.dataclass
class SequenceLength(Metric):
  """Metric for length for a sequence example.

  Example::

    example = {'y': jnp.array([1, 2, 3, 4, 0, 0])}
    prediction = jnp.array([])  # Unused.
    metric = SequenceLength()
    print(metric.evaluate_example(example, prediction))
    # MeanStat(accum=4, weight=1) => 4

  Attributes:
    target_key: Key name in ``example`` for target.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
  """
  target_key: str = 'y'
  masked_target_values: Tuple[int, ...] = (0,)

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes non masked length for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for ``example`` of shape [num_classes]

    Returns:
      MeanStat for non masked length for a single sequence.
    """
    del prediction
    target = example[self.target_key]
    target_weight = get_target_weight(target, self.masked_target_values)
    return MeanStat.new(jnp.sum(target_weight), jnp.sum(jnp.any(target_weight)))


@dataclasses.dataclass
class PerDomainMetric(Metric):
  """Turns a base metric into one that groups results by domain.

  This is useful in algorithms such as AgnosticFedAvg.

  ``example`` is expected to contain a feature named
  :attr:`~PerDomainMetric.domain_id_key`, which stores the integer domain id in
  [0, num_domains). PerDomain accumulates :attr:`~PerDomainMetric.base` 's
  :class:`Stat` within each domain. If the base :class:`Metric` returns a
  :class:`Stat` whose result is of shape X, then the :class:`Stat` returned by
  PerDomain will produce a result of shape ``(num_domains,) + X``.
  See :ref:`batching-stats` for the higher rank :class:`Stat` mechanism enabling
  this.

  Example::

    per_domain_accuracy = PerDomain(metrics.Accuracy(), num_domains=3)
    batch_example = {
        'domain_id': jnp.array([0, 0, 1, 2]),
        'y': jnp.array([0, 1, 0, 1])
    }
    batch_prediction = jnp.array([[0., 1.], [2., 3.], [4., 5.], [6., 7.]])
    print(
        evaluate_batch(per_domain_accuracy, batch_example,
                       batch_prediction).result())
    # [0.5 0.  1. ]
  """
  base: Metric
  num_domains: int
  domain_id_key: str = 'domain_id'

  def zero(self) -> Stat:

    def broadcast_to(x):
      return jnp.broadcast_to(x, (self.num_domains,) + x.shape)

    return jax.tree_util.tree_map(broadcast_to, self.base.zero())

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> Stat:
    domain_mask = jax.nn.one_hot(
        example[self.domain_id_key], self.num_domains, dtype=jnp.bool_)

    def where(a, b):
      return apply_mask(domain_mask, jnp.expand_dims(a, 0),
                        jnp.expand_dims(b, 0))

    return jax.tree_util.tree_multimap(
        where, self.base.evaluate_example(example, prediction),
        self.base.zero())


@dataclasses.dataclass
class ConfusionMatrix(Metric):
  """Metric for making a Confusion Matrix.

  A confusion matrix is an nxn matrix often used to describe the performance
  of a classification model on a set of test data for which the true values are
  known. The model's predictions are represented through the columns, and the
  known data values through the rows. This allows one to view in which areas the
  model is doing well, as well as where there is room for improvement. For each
  row in the confusion matrix, if there are a lot of numbers outside of the main
  diagonal, the model is not doing so well in respect to when it is supposed to
  output that row's relative output class.

  Theoretical Example:
                Predicted P     Predicted N

    Actual P       TP               FN

    Actual N       FP               TN

    **This is for a binary classification model but the same concept applies
    to any model with n outputs. Notice that the TPs and TNs will always
    lie in the main diagonal of the matrix.

  Example::

    example = {'y': jnp.array(2)}
    prediction = jnp.array([0., 1., 0.])
    metric = ConfusionMatrix(num_classes=3)
    print(metric.evaluate_example(example, prediction))
    # SumStat(accum=DeviceArray([[0., 0., 0.],
                                 [0., 0., 0.],
                                 [0., 1., 0.]], dtype=float32)) => [[0. 0. 0.]
                                                                    [0. 0. 0.]
                                                                    [0. 1. 0.]]

  Attributes:
    target_key: Key name in ``example`` for target.
    pred_key: Key name in ``prediction`` for unnormalized model output pred.
    num_classes: Number of output classes of the model. Used to generate a
      matrix of shape [num_classes, num_classes].
  """
  num_classes: int
  target_key: str = 'y'
  pred_key: Optional[str] = None

  def zero(self) -> SumStat:
    return SumStat.new(jnp.zeros((self.num_classes, self.num_classes)))

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> SumStat:
    """Computes a Confusion Matrix for a single example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for ``example`` of shape [num_classes]

    Returns:
      SumStat for the Confusion Matrix for a single example.

    Raises:
      ValueError: If the num_classes attribute is not equal to the number of
        output classes of the model.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    if self.num_classes != len(pred):
      raise ValueError('Make sure num_classes is equal to the number of output '
                       f'classes of the model. num_classes: {self.num_classes} '
                       f'number of output classes of the model: {len(pred)}')
    pred_idx = jnp.argmax(pred, axis=-1)
    confusion_matrix = jnp.zeros((self.num_classes, self.num_classes))
    return SumStat.new(confusion_matrix.at[target, pred_idx].set(1))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Slimmed down dataclass."""

import dataclasses
import jax


# Forked and slimmed down from
# https://flax.readthedocs.io/en/latest/_modules/flax/struct.html#dataclass
# https://github.com/google/jax/issues/2371
def dataclass(clz: type):
  """Creates a dataclass which can be passed to functional transformations."""
  data_clz = dataclasses.dataclass(frozen=True)(clz)
  meta_fields = []
  data_fields = []
  for name, field_info in data_clz.__dataclass_fields__.items():
    is_pytree_node = field_info.metadata.get('pytree_node', True)
    if is_pytree_node:
      data_fields.append(name)
    else:
      meta_fields.append(name)

  def replace(self, **updates):
    """"Returns a new object replacing the specified fields with new values."""
    return dataclasses.replace(self, **updates)

  data_clz.replace = replace

  def iterate_clz(x):
    meta = tuple(getattr(x, name) for name in meta_fields)
    data = tuple(getattr(x, name) for name in data_fields)
    return data, meta

  def clz_from_iterable(meta, data):
    meta_args = tuple(zip(meta_fields, meta))
    data_args = tuple(zip(data_fields, data))
    kwargs = dict(meta_args + data_args)
    return data_clz(**kwargs)

  jax.tree_util.register_pytree_node(data_clz, iterate_clz, clz_from_iterable)
  return data_clz



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library for defining regularizers."""

import abc
from typing import Optional

from fedjax.legacy.core.typing import Params
import jax
import jax.numpy as jnp


class Regularizer(metaclass=abc.ABCMeta):
  """Base class representing regularizers."""

  def __init__(self, center_params: Optional[Params]):
    self._center_params = center_params

    def center_fn(params):
      return jax.tree_multimap(lambda a, b: a - b, params, center_params)

    if center_params is not None:
      self._preprocess_fn = center_fn
    else:
      self._preprocess_fn = lambda p: p

  @abc.abstractmethod
  def __call__(self, params: Params) -> float:
    """Evaluates the regularizer."""


class L2Regularizer(Regularizer):
  """Class representing an L2 regularizer.

  Attributes:
    _center_params: Param values to regularize toward.
    _weight: Weight in front of L2 norm.
    _param_weights: Optional per-parameter weighting, which allows different
      regularization strength for each parameter.
  """

  def __init__(
      self,
      center_params: Optional[Params] = None,
      weight: float = 1.0,
      param_weights: Optional[Params] = None):
    super().__init__(center_params)
    self._weight = weight
    self._param_weights = param_weights

  def __call__(self, params: Params) -> float:
    """Evaluates the regularizer."""
    params = self._preprocess_fn(params)
    leaves, _ = jax.tree_flatten(params)

    if self._param_weights:
      param_weight_leaves, _ = jax.tree_flatten(self._param_weights)
      return sum(jnp.vdot(pw, jnp.square(x))
                 for pw, x in zip(param_weight_leaves, leaves)) * self._weight

    return sum(jnp.vdot(x, x) for x in leaves) * self._weight



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated stackoverflow."""

import itertools
from typing import Callable, List, Optional, Tuple

from fedjax.core import client_datasets
from fedjax.core import federated_data
from fedjax.core import sqlite_federated_data
from fedjax.datasets import downloads
import numpy as np
import tensorflow as tf

SPLITS = ('train', 'held_out', 'test')


def cite():
  return """@misc{stackoverflow2019,
 title={TensorFlow Federated Stack Overflow dataset},
  author={The TensorFlow Federated Authors.},
  year={2019},
}"""


def load_split(split: str,
               mode: str = 'sqlite',
               cache_dir: Optional[str] = None) -> federated_data.FederatedData:
  """Loads a stackoverflow split.

  All bytes arrays are stored with `dtype=np.object`.

  Features:
    creation_date: [N] bytes array. Textual timestamp, e.g.
      b'2018-02-28 19:06:18.34 UTC'.
    title: [N] bytes array. The title of a post.
    score: [N] int64 array. The score of a post.
    tags: [N] bytes array. '|' separated list of tags, e.g. b'mysql|join'.
    tokens: [N] bytes array. Space separated list of tokens.
    type: [N] bytes array. Either b'question' or b'answer'.

  Args:
    split: Name of the split. One of SPLITS.
    mode: 'sqlite'.
    cache_dir: Directory to cache files in 'sqlite' mode.

  Returns:
    FederatedData.
  """
  if split not in SPLITS:
    raise ValueError(f'Invalid split={split!r}')
  if cache_dir is not None and mode != 'sqlite':
    raise ValueError('Caching locally is only supported in "sqlite" mode')
  if mode == 'sqlite':
    path = downloads.maybe_download(
        f'https://storage.googleapis.com/gresearch/fedjax/stackoverflow/stackoverflow_{split}.sqlite',
        cache_dir)
    return sqlite_federated_data.SQLiteFederatedData.new(path)
  else:
    raise ValueError(f'Unsupported mode={mode!r}')


def load_data(
    mode: str = 'sqlite',
    cache_dir: Optional[str] = None
) -> Tuple[federated_data.FederatedData, federated_data.FederatedData,
           federated_data.FederatedData]:
  """Loads partially preprocessed stackoverflow splits.

  Features:
    domain_id: [N] int32 domain id derived from type (question = 0; answer = 1).
    tokens: [N] bytes array. Space separated list of tokens.

  To convert `tokens` into padded/truncated integer labels, use a
  StackoverflowTokenizer. For example,

  ```
  from fedjax.core.datasets import stackoverflow
  # Load partially preprocessed splits.
  train, held_out, test = stackoverflow.load_data()
  # Apply tokenizer during batching.
  tokenizer = stackoverflow.StackoverflowTokenizer()
  train_max_length, eval_max_length = 20, 30
  train_for_train = train.preprocess_batch(
      tokenizer.as_preprocess_batch(train_max_length))
  train_for_eval = train.preprocess_batch(
      tokenizer.as_preprocess_batch(eval_max_length))
  held_out = held_out.preprocess_batch(
      tokenizer.as_preprocess_batch(eval_max_length))
  test = test.preprocess_batch(
      tokenizer.as_preprocess_batch(eval_max_length))
  ```

  Features after tokenization:
    domain_id: Same as before.
    x: [N, max_length] int32 array of padded/truncated input labels.
    y: [N, max_length] int32 array of padded/truncated output labels.

  Args:
    mode: 'sqlite'.
    cache_dir: Directory to cache files in 'sqlite' mode.

  Returns:
    A (train, held_out, test) tuple of federated data.
  """
  train = load_split('train', mode, cache_dir)
  held_out = load_split('held_out', mode, cache_dir)
  test = load_split('test', mode, cache_dir)
  return (train.preprocess_client(preprocess_client),
          held_out.preprocess_client(preprocess_client),
          test.preprocess_client(preprocess_client))


def preprocess_client(
    client_id: federated_data.ClientId,
    examples: client_datasets.Examples) -> client_datasets.Examples:
  """Attaches domain id and drops features other than `tokens`."""
  del client_id
  return {
      'domain_id': (examples['type'] == b'answer').astype(np.int32),
      'tokens': examples['tokens']
  }


def default_vocab(default_vocab_size) -> List[str]:
  """Loads the deafult stackoverflow vocabulary."""
  path = 'gs://gresearch/fedjax/stackoverflow/stackoverflow.word_count'
  vocab = []
  with tf.io.gfile.GFile(path) as f:
    for line in itertools.islice(f, default_vocab_size):
      word, _ = line.split()
      vocab.append(word)
  return vocab


# TODO(wuke): Remove dependency on TensorFlow.
class StackoverflowTokenizer:
  """Tokenizer for the `tokens` feature in stackoverflow.

  See load_data() for examples.
  """
  PAD = 0
  BOS = 1
  EOS = 2

  def __init__(self,
               vocab: Optional[List[str]] = None,
               default_vocab_size: Optional[int] = 10000,
               num_oov_buckets: int = 1):
    """Initializes a tokenizer.

    Args:
      vocab: Optional vocabulary. If specified, `default_vocab_size` is ignored.
        If None, `default_vocab_size` is used to load the standard vocabulary.
        This vocabulary should NOT have special tokens PAD, EOS, BOS, and OOV.
        The special tokens are added and handled automatically by the tokenizer.
        The preprocessed examples will have vocabulary size `len(vocab) + 3 +
        num_oov_buckets`.
      default_vocab_size: Number of words in the default vocabulary. This is
        only used when `vocab` is not specified. The preprocessed examples will
        have vocabulary size `default_vocab_size + 3 + num_oov_buckets`
        with 3 special labels: 0 (PAD), 1 (BOS), 2 (EOS), and `num_oov_buckets`
        OOV labels starting at `default_vocab_size + 3`.
      num_oov_buckets: Number of out of vocabulary buckets.
    """
    if vocab is None:
      # Load default vocabulary.
      vocab = default_vocab(default_vocab_size)
    with tf.device('cpu'):
      self._table = tf.lookup.StaticVocabularyTable(
          tf.lookup.KeyValueTensorInitializer(
              vocab, tf.range(len(vocab), dtype=tf.int64)),
          num_oov_buckets=num_oov_buckets)

  def as_preprocess_batch(
      self, max_length: int
  ) -> Callable[[client_datasets.Examples], client_datasets.Examples]:
    """Creates a preprocess_batch function.

    Args:
      max_length: The length to pad x/y sequences to. Sequences longer than this
        are also truncated to this length.

    Returns:
      A function that can be used with FederatedData.preprocess_batch().
    """

    @tf.function(input_signature=[tf.TensorSpec(shape=[None], dtype=tf.string)])
    def token_to_ids(tokens):
      """Tokenizes tokens with TensorFlow ops."""
      words = tf.strings.split(tokens, sep=' ')
      # Add 3 to reserve 0 (pad), 1 (bos), and 2 (eos). Hence offseting other
      # ids by 3.
      token_ids = self._table.lookup(words) + 3
      # Pad bos/eos.
      batch_bos = tf.zeros_like(
          tokens[..., tf.newaxis], dtype=token_ids.dtype) + self.BOS
      batch_eos = tf.zeros_like(
          tokens[..., tf.newaxis], dtype=token_ids.dtype) + self.EOS
      token_ids = tf.concat([batch_bos, token_ids, batch_eos], axis=-1)
      # Turn into x and y
      x = tf.cast(token_ids[..., :-1], tf.int32)
      y = tf.cast(token_ids[..., 1:], tf.int32)
      # Turn into dense tensors.
      shape = tokens.shape + [max_length]
      return (x.to_tensor(self.PAD,
                          shape=shape), y.to_tensor(self.PAD, shape=shape))

    def preprocess_batch(
        examples: client_datasets.Examples) -> client_datasets.Examples:
      with tf.device('cpu'):
        x, y = token_to_ids(examples['tokens'])
      result = {'x': x.numpy(), 'y': y.numpy()}
      if 'domain_id' in examples:
        result['domain_id'] = examples['domain_id']
      return result

    return preprocess_batch



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Toy data and model for testing algorithms."""

import collections

from typing import List, Tuple

from fedjax.legacy.core import federated_algorithm
from fedjax.legacy.core import metrics
from fedjax.legacy.core import model
from fedjax.legacy.core.typing import FederatedData
import haiku as hk
import numpy as np
import optax
import tensorflow as tf
import tensorflow_federated as tff


def create_toy_data(num_clients: int, num_clusters: int, num_classes: int,
                    num_examples: int, seed: int) -> FederatedData:
  r"""Creates toy density estimation federated data.

  This is a synthetic dataset for density estimation. Let X=∅, Y=[d] where d
  is
  the number of classes. Let l be cross entropy loss and the number of users p.
  We create client distributions as a mixture with a uniform component, a
  cluster component, and an individual component. The details of the
  distributions are below.

  Let U be a uniform distribution over Y.  Let P_k be a point mass
  distribution with P_k(k) = 1.0 and for all y != k, P_k(y) = 0.
  For a client k, let D_k be a mixture given by

  D_k = 0.5 * P_{k%4} + 0.25 * U + 0.25 * P_{k%(d-4)}

  where a%b, is a modulo b. Roughly, U is the uniform component and is the same
  for all the clients, P_{k%4} is the cluster component and same for clients
  with same clusters, and P_{k%(d-4)} is the individual component per client.

  Args:
    num_clients: An int number of clients to generate.
    num_clusters: An int number of clusters to generate.
    num_classes: An int number of output classes.
    num_examples: An int number of examples per client.
    seed: An int random seed for sampling.

  Returns:
    Federated data containing randomly generated examples of structure:
      x: A numpy array of ones of shape [1].
      y: A numpy array of shape [].
  """
  distributions = 0.5 * np.identity(num_clusters)
  zeros = np.zeros((num_clusters, num_classes - num_clusters))
  distributions = np.concatenate([distributions, zeros], axis=-1)
  distributions += 0.25 * np.ones((num_clusters, num_classes)) / num_classes

  random_state = np.random.RandomState(seed)
  ids = random_state.permutation(num_clients) % num_clusters
  ids = np.arange(num_clients) % num_clusters

  client_data = collections.OrderedDict()
  for i in range(num_clients):
    local_change = np.zeros((num_classes))
    local_change[num_clusters + (i % (num_classes - num_clusters - 1)) +
                 1] = 0.25
    distribution = distributions[ids[i], :] + local_change
    client_id = str(i)
    client_data[client_id] = random_state.choice(
        num_classes, num_examples, p=distribution)

  for client_id, client_points in client_data.items():
    client_data[client_id] = collections.OrderedDict(
        x=tf.expand_dims(np.ones_like(client_points), axis=-1),
        y=client_points,
    )

  return tff.simulation.datasets.TestClientData(client_data)


def create_toy_model(num_classes: int,
                     num_hidden_layers: int = 0,
                     feature_dim: int = 1) -> model.Model:
  """Creates toy haiku model.

  Args:
    num_classes: Number of output classes.
    num_hidden_layers: Number of hidden layers.
    feature_dim: Input feature dimension.

  Returns:
    Toy haiku model.
  """

  def forward_pass(batch):
    network = hk.Sequential(
        [hk.Linear(2 * num_classes) for _ in range(num_hidden_layers)] +
        [hk.Linear(num_classes)])
    return network(batch['x'])

  transformed_forward_pass = hk.transform(forward_pass)
  sample_batch = collections.OrderedDict(
      x=np.zeros((1, feature_dim)), y=np.zeros((1,)))

  def loss(batch, preds):
    return metrics.cross_entropy_loss_fn(targets=batch['y'], preds=preds)

  def accuracy(batch, preds):
    return metrics.accuracy_fn(targets=batch['y'], preds=preds)

  metrics_fn_map = collections.OrderedDict(accuracy=accuracy)
  return model.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=sample_batch,
      loss_fn=loss,
      metrics_fn_map=metrics_fn_map)


def create_toy_example(num_clients: int = 10,
                       num_clusters: int = 2,
                       num_classes: int = 4,
                       num_examples: int = 5,
                       num_hidden_layers: int = 0,
                       seed: int = 1) -> Tuple[FederatedData, model.Model]:
  """Creates toy data and model.

  Args:
    num_clients: Number of clients to generate.
    num_clusters: Number of clusters to generate.
    num_classes: Number of output classes.
    num_examples: Number of examples per client.
    num_hidden_layers: Number of hidden layers.
    seed: Random seed for sampling.

  Returns:
    A tuple of randomly generated federated toy data and toy model.
  """
  toy_data = create_toy_data(num_clients, num_clusters, num_classes,
                             num_examples, seed)
  toy_model = create_toy_model(num_classes, num_hidden_layers, feature_dim=1)
  return toy_data, toy_model


MockState = collections.namedtuple('MockState',
                                   ['params', 'opt_state', 'count'])


def create_mock_state(seed: int = 0) -> MockState:
  """Returns example nested struct with model parameters and optimizer state."""

  def net_fn(x):
    network = hk.Sequential([
        hk.Flatten(),
        hk.Linear(62),
    ])
    return network(x)

  transformed = hk.transform(net_fn)
  rng_seq = hk.PRNGSequence(seed)
  params = transformed.init(next(rng_seq), np.zeros((1, 1)))
  opt_init, _ = optax.sgd(learning_rate=1.0, momentum=0.9)
  opt_state = opt_init(params)
  return MockState(params=params, opt_state=opt_state, count=0)


class MockFederatedAlgorithm(federated_algorithm.FederatedAlgorithm):
  """Mock federated algorithm implementation."""

  def __init__(self, *args, **kwargs):
    self._federated_data, self._model = create_toy_example(*args, **kwargs)
    self._rng_seq = hk.PRNGSequence(0)

  @property
  def federated_data(self) -> FederatedData:
    return self._federated_data

  @property
  def model(self) -> model.Model:
    return self._model

  def init_state(self) -> MockState:
    """Returns initial state of algorithm."""
    params = self._model.init_params(next(self._rng_seq))
    opt_state = optax.sgd(learning_rate=1.0, momentum=0.9)[0](params)
    count = 0
    return MockState(params, opt_state, count)

  def run_round(self, state: MockState, client_ids: List[str]) -> MockState:
    """Runs one round of federated training."""
    count = state.count
    for client_id in client_ids:
      for _ in self._federated_data.create_tf_dataset_for_client(client_id):
        count += 1
    return MockState(
        params=state.params, opt_state=state.opt_state, count=count)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.regularizers."""

from absl.testing import absltest

from fedjax.core import regularizers

import jax
import jax.numpy as jnp


class RegularizersTest(absltest.TestCase):

  def test_l2_regularizer(self):
    params = {
        'linear0': {
            'w': jnp.array([1., 2., 3.])
        },
        'linear1': {
            'w': jnp.array([4., 5., 6.]),
            'b': jnp.array([1., 1., 1.])
        }
    }

    with self.subTest('weight'):
      regularizer = regularizers.l2_regularizer(weight=0.2)
      self.assertAlmostEqual(regularizer(params), 0.2 * 94, places=5)

    with self.subTest('center_params'):
      center_params = jax.tree_util.tree_map(jnp.ones_like, params)
      regularizer = regularizers.l2_regularizer(
          weight=0.2, center_params=center_params)
      self.assertAlmostEqual(regularizer(params), 0.2 * 55)

    with self.subTest('params_weights'):
      params_weights = jax.tree_util.tree_map(lambda l: jnp.ones_like(l) * 2,
                                              params)
      regularizer = regularizers.l2_regularizer(
          weight=0.2, params_weights=params_weights)
      self.assertAlmostEqual(regularizer(params), 0.2 * 188, places=5)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Stack Overflow recurrent models."""

from typing import Optional

from fedjax.core import metrics
from fedjax.core import models

import haiku as hk
import jax.numpy as jnp


def create_lstm_model(vocab_size: int = 10000,
                      embed_size: int = 96,
                      lstm_hidden_size: int = 670,
                      lstm_num_layers: int = 1,
                      share_input_output_embeddings: bool = False,
                      expected_length: Optional[float] = None) -> models.Model:
  """Creates LSTM language model.

  Word-level language model for Stack Overflow.
  Defaults to the model used in:

  Adaptive Federated Optimization
    Sashank Reddi, Zachary Charles, Manzil Zaheer, Zachary Garrett, Keith Rush,
    Jakub Konečný, Sanjiv Kumar, H. Brendan McMahan.
    https://arxiv.org/abs/2003.00295

  Args:
    vocab_size: The number of possible output words. This does not include
      special tokens like PAD, BOS, EOS, or OOV.
    embed_size: Embedding size for each word.
    lstm_hidden_size: Hidden size for LSTM cells.
    lstm_num_layers: Number of LSTM layers.
    share_input_output_embeddings: Whether to share the input embeddings with
      the output logits.
    expected_length: Expected average sentence length used to scale the training
      loss down by `1. / expected_length`. This constant term is used so that
      the total loss over all the words in a sentence can be scaled down to per
      word cross entropy values by a constant factor instead of dividing by
      number of words which can vary across batches. Defaults to no scaling.

  Returns:
    Model.
  """
  # TODO(jaero): Replace these with direct references from dataset.
  pad = 0
  bos = 1
  eos = 2
  oov = vocab_size + 3
  full_vocab_size = vocab_size + 4
  # We do not guess EOS, and if we guess OOV, it's treated as a mistake.
  logits_mask = [0. for _ in range(full_vocab_size)]
  for i in (pad, bos, eos, oov):
    logits_mask[i] = jnp.NINF
  logits_mask = tuple(logits_mask)

  def forward_pass(batch):
    x = batch['x']
    # [time_steps, batch_size, ...].
    x = jnp.transpose(x)
    # [time_steps, batch_size, embed_dim].
    embedding_layer = hk.Embed(full_vocab_size, embed_size)
    embeddings = embedding_layer(x)

    lstm_layers = []
    for _ in range(lstm_num_layers):
      lstm_layers.extend([
          hk.LSTM(hidden_size=lstm_hidden_size),
          jnp.tanh,
          # Projection changes dimension from lstm_hidden_size to embed_size.
          hk.Linear(embed_size)
      ])
    rnn_core = hk.DeepRNN(lstm_layers)
    initial_state = rnn_core.initial_state(batch_size=embeddings.shape[1])
    # [time_steps, batch_size, hidden_size].
    output, _ = hk.static_unroll(rnn_core, embeddings, initial_state)

    if share_input_output_embeddings:
      output = jnp.dot(output, jnp.transpose(embedding_layer.embeddings))
      output = hk.Bias(bias_dims=[-1])(output)
    else:
      output = hk.Linear(full_vocab_size)(output)
    # [batch_size, time_steps, full_vocab_size].
    output = jnp.transpose(output, axes=(1, 0, 2))
    return output

  def train_loss(batch, preds):
    """Returns total loss per sentence optionally scaled down to token level."""
    targets = batch['y']
    per_token_loss = metrics.unreduced_cross_entropy_loss(targets, preds)
    # Don't count padded values in loss.
    per_token_loss *= targets != pad
    sentence_loss = jnp.sum(per_token_loss, axis=-1)
    if expected_length is not None:
      return sentence_loss * (1. / expected_length)
    return sentence_loss

  transformed_forward_pass = hk.transform(forward_pass)
  return models.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch={
          'x': jnp.zeros((1, 1), dtype=jnp.int32),
          'y': jnp.zeros((1, 1), dtype=jnp.int32),
      },
      train_loss=train_loss,
      eval_metrics={
          'accuracy_in_vocab':
              metrics.SequenceTokenAccuracy(
                  masked_target_values=(pad, eos), logits_mask=logits_mask),
          'accuracy_no_eos':
              metrics.SequenceTokenAccuracy(masked_target_values=(pad, eos)),
          'num_tokens':
              metrics.SequenceTokenCount(masked_target_values=(pad,)),
          'sequence_length':
              metrics.SequenceLength(masked_target_values=(pad,)),
          'sequence_loss':
              metrics.SequenceCrossEntropyLoss(masked_target_values=(pad,)),
          'token_loss':
              metrics.SequenceTokenCrossEntropyLoss(
                  masked_target_values=(pad,)),
          'token_oov_rate':
              metrics.SequenceTokenOOVRate(
                  oov_target_values=(oov,), masked_target_values=(pad,)),
          'truncation_rate':
              metrics.SequenceTruncationRate(
                  eos_target_value=eos, masked_target_values=(pad,)),
      })



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""FedJAX API."""

# Public core API.
from fedjax.core import client_samplers
from fedjax.core import metrics
from fedjax.core import optimizers
from fedjax.core import regularizers
from fedjax.core import serialization
from fedjax.core import tree_util

from fedjax.core.client_datasets import BatchHParams
from fedjax.core.client_datasets import BatchPreprocessor
from fedjax.core.client_datasets import buffered_shuffle_batch_client_datasets
from fedjax.core.client_datasets import ClientDataset
from fedjax.core.client_datasets import EXAMPLE_MASK_KEY
from fedjax.core.client_datasets import padded_batch_client_datasets
from fedjax.core.client_datasets import PaddedBatchHParams
from fedjax.core.client_datasets import ShuffleRepeatBatchHParams

from fedjax.core.dataclasses import dataclass

from fedjax.core.federated_algorithm import FederatedAlgorithm

from fedjax.core.federated_data import ClientPreprocessor
from fedjax.core.federated_data import FederatedData
from fedjax.core.federated_data import padded_batch_federated_data
from fedjax.core.federated_data import RepeatableIterator
from fedjax.core.federated_data import shuffle_repeat_batch_federated_data
from fedjax.core.federated_data import SubsetFederatedData

from fedjax.core.for_each_client import for_each_client
from fedjax.core.for_each_client import for_each_client_backend
from fedjax.core.for_each_client import get_for_each_client_backend
from fedjax.core.for_each_client import set_for_each_client_backend
from fedjax.core.for_each_client import ForEachClientBackend

from fedjax.core.models import AverageLossEvaluator
from fedjax.core.models import create_model_from_haiku
from fedjax.core.models import create_model_from_stax
from fedjax.core.models import evaluate_average_loss
from fedjax.core.models import evaluate_model
from fedjax.core.models import grad
from fedjax.core.models import Model
from fedjax.core.models import model_grad
from fedjax.core.models import model_per_example_loss
from fedjax.core.models import ModelEvaluator

from fedjax.core.sqlite_federated_data import SQLiteFederatedData
from fedjax.core.sqlite_federated_data import SQLiteFederatedDataBuilder

from fedjax.core.in_memory_federated_data import InMemoryFederatedData

from fedjax.core.optimizers import Optimizer
from fedjax.core.typing import BatchExample
from fedjax.core.typing import BatchPrediction
from fedjax.core.typing import OptState
from fedjax.core.typing import Params
from fedjax.core.typing import PRNGKey
from fedjax.core.typing import PyTree
from fedjax.core.typing import SingleExample
from fedjax.core.typing import SinglePrediction

# Subpackages available by default.
from fedjax import aggregators
from fedjax import algorithms
from fedjax import datasets
from fedjax import models
from fedjax import training

# Please do not use symbols in `core` as they are not part of the public API.
try:
  del core  # pylint: disable=undefined-variable
except NameError:
  pass



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Structured flags commonly used in experiment binaries.

Structured flags are often used to construct complex structures via multiple
simple flags (e.g. an optimizer can be created by controlling learning rate and
other hyper parameters).
"""

import sys
from typing import Optional, Sequence

from absl import flags
from fedjax.core import client_datasets
from fedjax.core import optimizers
from fedjax.training import federated_experiment
from fedjax.training import tasks

FLAGS = flags.FLAGS


class NamedFlags:
  """A group of flags with an optional named prefix."""

  def __init__(self, name: Optional[str]):
    self._name = name
    if name is None:
      self._prefix = ''
      self._description = ''
    else:
      self._prefix = name + '_'
      self._description = f' for {name}'

  # Two special things:
  # - Add prefix/description when registering flags.
  # - Setting module_name so that these flags become visible for -help.

  def _enum(self, name: str, default: Optional[str], enum_values: Sequence[str],
            help_: str):
    flags.DEFINE_enum(
        self._prefix + name,
        default,
        enum_values,
        help_ + self._description,
        module_name=sys.argv[0])

  def _string(self, name: str, default: Optional[str], help_: str):
    flags.DEFINE_string(
        self._prefix + name,
        default,
        help_ + self._description,
        module_name=sys.argv[0])

  def _float(self, name: str, default: Optional[float], help_: str):
    flags.DEFINE_float(
        self._prefix + name,
        default,
        help_ + self._description,
        module_name=sys.argv[0])

  def _integer(self, name: str, default: Optional[int], help_: str):
    flags.DEFINE_integer(
        self._prefix + name,
        default,
        help_ + self._description,
        module_name=sys.argv[0])

  def _get_flag(self, flag):
    return getattr(FLAGS, self._prefix + flag)


class OptimizerFlags(NamedFlags):
  """Constructs a fedjax.Optimizer from flags."""

  SUPPORTED = ('sgd', 'momentum', 'adam', 'rmsprop', 'adagrad')

  def __init__(self,
               name: Optional[str] = None,
               default_optimizer: str = 'sgd'):
    super().__init__(name)
    self._enum('optimizer', default_optimizer, self.SUPPORTED, 'Optimizer')
    self._float('learning_rate', 0.005, 'Server step size')
    # Momentum parameters.
    self._float('momentum', 0.0, 'Momentum parameter')
    # Adam parameters
    self._float('adam_beta1', 0.9, 'Adam beta 1 parameter')
    self._float('adam_beta2', 0.99, 'Adam beta 2 parameter')
    self._float('adam_epsilon', 1e-3, 'Adam epsilon parameter')
    # RMSprop parameters.
    self._float('rmsprop_decay', 0.9, 'RMSProp decay parameter')
    self._float('rmsprop_epsilon', 1e-3, 'RMSprop epsilon parameter')
    # Adagrad parameters.
    self._float(
        'adagrad_epsilon', 1e-6,
        'Adagrad epsilon parameter that is added to second moment' +
        self._description)

  def get(self) -> optimizers.Optimizer:
    """Gets the specified optimizer."""
    optimizer_name = self._get_flag('optimizer')
    learning_rate = self._get_flag('learning_rate')
    if optimizer_name == 'sgd':
      return optimizers.sgd(learning_rate)
    elif optimizer_name == 'momentum':
      return optimizers.sgd(learning_rate, self._get_flag('momentum'))
    elif optimizer_name == 'adam':
      return optimizers.adam(learning_rate, self._get_flag('adam_beta1'),
                             self._get_flag('adam_beta2'),
                             self._get_flag('adam_epsilon'))
    elif optimizer_name == 'rmsprop':
      return optimizers.rmsprop(learning_rate, self._get_flag('rmsprop_decay'),
                                self._get_flag('rmsprop_epsilon'))
    elif optimizer_name == 'adagrad':
      return optimizers.adagrad(
          learning_rate, eps=self._get_flag('adagrad_epsilon'))
    else:
      raise ValueError(f'Unsupported optimizer {optimizer_name!r} from '
                       f'--{self._prefix}optimizer.')


class ShuffleRepeatBatchHParamsFlags(NamedFlags):
  """Constructs ShuffleRepeatBatchHParams from flags."""

  def __init__(self, name: Optional[str] = None, default_batch_size: int = 128):
    super().__init__(name)
    defaults = client_datasets.ShuffleRepeatBatchHParams(batch_size=-1)
    # TODO(wuke): Support other fields.
    self._integer('batch_size', default_batch_size, 'Batch size')
    self._integer('num_epochs', defaults.num_epochs, 'Number of epochs')
    self._integer('num_steps', defaults.num_steps, 'Number of steps')

  def get(self):
    return client_datasets.ShuffleRepeatBatchHParams(
        batch_size=self._get_flag('batch_size'),
        num_epochs=self._get_flag('num_epochs'),
        num_steps=self._get_flag('num_steps'))


class PaddedBatchHParamsFlags(NamedFlags):
  """Constructs PaddedBatchHParams from flags."""

  def __init__(self, name: Optional[str] = None, default_batch_size: int = 128):
    super().__init__(name)
    # TODO(wuke): Support other fields.
    self._integer('batch_size', default_batch_size, 'Batch size')

  def get(self):
    return client_datasets.PaddedBatchHParams(
        batch_size=self._get_flag('batch_size'))


class BatchHParamsFlags(NamedFlags):
  """Constructs BatchHParams from flags."""

  def __init__(self, name: Optional[str] = None, default_batch_size: int = 128):
    super().__init__(name)
    # TODO(wuke): Support other fields.
    self._integer('batch_size', default_batch_size, 'Batch size')

  def get(self):
    return client_datasets.BatchHParams(batch_size=self._get_flag('batch_size'))


class FederatedExperimentConfigFlags(NamedFlags):
  """Constructs FederatedExperimentConfig from flags."""

  def __init__(self, name: Optional[str] = None):
    super().__init__(name)
    defaults = federated_experiment.FederatedExperimentConfig(
        root_dir='', num_rounds=-1)
    self._string('root_dir', None, 'Root directory of experiment outputs')
    self._integer('num_rounds', None, 'Number of federated training rounds')
    self._integer(
        'checkpoint_frequency', defaults.checkpoint_frequency,
        'Checkpoint frequency in rounds' +
        '. If <= 0, no checkpointing is done.')
    self._integer('num_checkpoints_to_keep', defaults.num_checkpoints_to_keep,
                  'Maximum number of checkpoints to keep')
    self._integer(
        'eval_frequency', defaults.eval_frequency,
        'Evaluation frequency in rounds' + '. If <= 0, no evaluation is done.')

  def get(self):
    return federated_experiment.FederatedExperimentConfig(
        root_dir=self._get_flag('root_dir'),
        num_rounds=self._get_flag('num_rounds'),
        checkpoint_frequency=self._get_flag('checkpoint_frequency'),
        num_checkpoints_to_keep=self._get_flag('num_checkpoints_to_keep'),
        eval_frequency=self._get_flag('eval_frequency'))


class TaskFlags(NamedFlags):
  """Constructs a standard task tuple from flags."""

  def __init__(self, name: Optional[str] = None):
    super().__init__(name)
    self._enum('task', None, tasks.ALL_TASKS, 'Which task to run')
    self._string('data_mode', 'sqlite', 'Data loading mode')
    self._string('cache_dir', None,
                 'Cache directory when loading SQLite federated data')

  def get(self):
    return tasks.get_task(
        self._get_flag('task'), self._get_flag('data_mode'),
        self._get_flag('cache_dir'))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for hyp_cluster."""

from absl.testing import absltest
from fedjax.algorithms import hyp_cluster
from fedjax.core import client_datasets
from fedjax.core import metrics
from fedjax.core import models
from fedjax.core import optimizers
import jax
import jax.numpy as jnp
import numpy as np
import numpy.testing as npt


class ClientTrainerTest(absltest.TestCase):

  OPTIMIZER = optimizers.sgd(1)

  def test_train_global_params(self):

    def grad(params, batch, rng):
      return 0.5 * params + jnp.mean(batch['x']) + jax.random.uniform(rng, [])

    rng_0 = jax.random.PRNGKey(0)
    rng_uniform_00 = jax.random.uniform(jax.random.split(rng_0)[1], [])
    rng_uniform_01 = jax.random.uniform(
        jax.random.split(jax.random.split(rng_0)[0])[1], [])
    rng_1 = jax.random.PRNGKey(1)
    rng_uniform_10 = jax.random.uniform(jax.random.split(rng_1)[1], [])

    params = jnp.array(1.)
    clients = [(b'0000', [{
        'x': jnp.array([0.125])
    }, {
        'x': jnp.array([0.25, 0.75])
    }], rng_0), (b'1001', [{
        'x': jnp.array([0, 1, 2])
    }], rng_1)]
    client_delta_params = dict(
        hyp_cluster.ClientDeltaTrainer(grad,
                                       self.OPTIMIZER).train_global_params(
                                           params, clients))
    self.assertCountEqual(client_delta_params, [b'0000', b'1001'])
    npt.assert_allclose(client_delta_params[b'0000'],
                        (0.5 * 1 + 0.125 + rng_uniform_00) +
                        (0.5 * (0.375 - rng_uniform_00) + 0.5 + rng_uniform_01))
    npt.assert_allclose(client_delta_params[b'1001'],
                        (0.5 * 1 + 1) + rng_uniform_10)

  def test_train_per_client_params(self):

    def grad(params, batch, rng):
      return 0.5 * params + jnp.mean(batch['x']) + jax.random.uniform(rng, [])

    rng_0 = jax.random.PRNGKey(0)
    rng_uniform_00 = jax.random.uniform(jax.random.split(rng_0)[1], [])
    rng_uniform_01 = jax.random.uniform(
        jax.random.split(jax.random.split(rng_0)[0])[1], [])
    rng_1 = jax.random.PRNGKey(1)
    rng_uniform_10 = jax.random.uniform(jax.random.split(rng_1)[1], [])

    clients = [
        (b'0000', [{
            'x': jnp.array([0.125])
        }, {
            'x': jnp.array([0.25, 0.75])
        }], rng_0, jnp.array(1.)),
        (b'1001', [{
            'x': jnp.array([0, 1, 2])
        }], rng_1, jnp.array(0.5)),
    ]
    client_delta_params = dict(
        hyp_cluster.ClientDeltaTrainer(
            grad, self.OPTIMIZER).train_per_client_params(clients))
    self.assertCountEqual(client_delta_params, [b'0000', b'1001'])
    npt.assert_allclose(client_delta_params[b'0000'],
                        (0.5 * 1 + 0.125 + rng_uniform_00) +
                        (0.5 * (0.375 - rng_uniform_00) + 0.5 + rng_uniform_01))
    npt.assert_allclose(client_delta_params[b'1001'],
                        (0.5 * 0.5 + 1) + rng_uniform_10)

  def test_return_params(self):

    def grad(params, batch, _):
      return 0.5 * params + jnp.mean(batch['x'])

    params = jnp.array(1.)
    clients = [(b'0000', [{
        'x': jnp.array([0.125])
    }, {
        'x': jnp.array([0.25, 0.75])
    }], jax.random.PRNGKey(0)),
               (b'1001', [{
                   'x': jnp.array([0, 1, 2])
               }], jax.random.PRNGKey(1))]
    client_delta_params = dict(
        hyp_cluster.ClientParamsTrainer(grad,
                                        self.OPTIMIZER).train_global_params(
                                            params, clients))
    npt.assert_equal(
        jax.device_get(client_delta_params), {
            b'0000': 1 - ((0.5 * 1 + 0.125) + (0.5 * 0.375 + 0.5)),
            b'1001': 1 - (0.5 * 1 + 1)
        })


class HypClusterTest(absltest.TestCase):

  def test_random_init(self):

    def init(rng):
      return jax.random.uniform(rng, [2])

    cluster_params = hyp_cluster.random_init(3, init, jax.random.PRNGKey(0))
    self.assertIsInstance(cluster_params, list)
    self.assertLen(cluster_params, 3)
    for i in cluster_params:
      self.assertIsInstance(i, jnp.ndarray)
      self.assertEqual(i.shape, (2,))
    self.assertTrue((cluster_params[0] != cluster_params[1]).any())
    self.assertTrue((cluster_params[0] != cluster_params[2]).any())
    self.assertTrue((cluster_params[1] != cluster_params[2]).any())

  def test_kmeans_init(self):
    functions_called = set()

    def init(rng):
      functions_called.add('init')
      return jax.random.uniform(rng)

    def apply_for_train(params, batch, rng):
      functions_called.add('apply_for_train')
      self.assertIsNotNone(rng)
      return params - batch['x']

    def train_loss(batch, out):
      functions_called.add('train_loss')
      return jnp.square(out) + batch['bias']

    def regularizer(params):
      del params
      functions_called.add('regularizer')
      return 0

    initializer = hyp_cluster.ModelKMeansInitializer(
        models.Model(
            init=init,
            apply_for_train=apply_for_train,
            apply_for_eval=None,
            train_loss=train_loss,
            eval_metrics={}), optimizers.sgd(0.5), regularizer)
    # Each client has 1 example, so it's very easy to reach minimal loss, at
    # which point the loss entirely depends on bias.
    clients = [
        (b'0',
         client_datasets.ClientDataset({
             'x': np.array([1.01]),
             'bias': np.array([-2.])
         }), jax.random.PRNGKey(1)),
        (b'1',
         client_datasets.ClientDataset({
             'x': np.array([3.02]),
             'bias': np.array([-1.])
         }), jax.random.PRNGKey(2)),
        (b'2',
         client_datasets.ClientDataset({
             'x': np.array([3.03]),
             'bias': np.array([1.])
         }), jax.random.PRNGKey(3)),
        (b'3',
         client_datasets.ClientDataset({
             'x': np.array([1.04]),
             'bias': np.array([2.])
         }), jax.random.PRNGKey(3)),
    ]
    train_batch_hparams = client_datasets.ShuffleRepeatBatchHParams(
        batch_size=1, num_epochs=5)
    eval_batch_hparams = client_datasets.PaddedBatchHParams(batch_size=2)
    # Using a rng that leads to b'0' being the initial center.
    cluster_params = initializer.cluster_params(
        num_clusters=3,
        rng=jax.random.PRNGKey(0),
        clients=clients,
        train_batch_hparams=train_batch_hparams,
        eval_batch_hparams=eval_batch_hparams)
    self.assertIsInstance(cluster_params, list)
    self.assertLen(cluster_params, 3)
    npt.assert_allclose(cluster_params, [1.01, 3.03, 1.04])
    self.assertCountEqual(
        functions_called,
        ['init', 'apply_for_train', 'train_loss', 'regularizer'])
    # Using a rng that leads to b'2' being the initial center.
    cluster_params = initializer.cluster_params(
        num_clusters=3,
        rng=jax.random.PRNGKey(1),
        clients=clients,
        train_batch_hparams=train_batch_hparams,
        eval_batch_hparams=eval_batch_hparams)
    self.assertIsInstance(cluster_params, list)
    self.assertLen(cluster_params, 3)
    npt.assert_allclose(cluster_params, [3.03, 1.04, 1.04])

  def test_maximization_step(self):
    # L1 distance from centers.
    def per_example_loss(params, batch, rng):
      # Randomly flip the center to test rng behavior.
      sign = jax.random.bernoulli(rng) * 2 - 1
      return jnp.abs(params * sign - batch['x'])

    def regularizer(params):
      return 0.01 * jnp.sum(jnp.abs(params))

    evaluator = models.AverageLossEvaluator(per_example_loss, regularizer)
    cluster_params = [jnp.array(0.), jnp.array(-1.), jnp.array(2.)]
    # Batch size is chosen so that we run 1 or 2 batches.
    batch_hparams = client_datasets.PaddedBatchHParams(batch_size=2)
    # Special seeds:
    # - No flip in first 2 steps for all 3 clusters: 0;
    # - Flip all in first 2 steps for all 3 clusters: 16;
    # - No flip then flip all for all 3 clusters: 68;
    # - Flips only cluster 1 in first 2 steps: 106.
    clients = [
        # No flip in first 2 steps for all 3 clusters.
        (b'near0', client_datasets.ClientDataset({'x': np.array([0.1])}),
         jax.random.PRNGKey(0)),
        # Flip all in first 2 steps for all 3 clusters.
        (b'near-1',
         client_datasets.ClientDataset({'x': np.array([0.9, 1.1, 1.3])}),
         jax.random.PRNGKey(16)),
        # No flip then flip all for all 3 clusters.
        (b'near2',
         client_datasets.ClientDataset({'x': np.array([1.9, 2.1, -2.1])}),
         jax.random.PRNGKey(68)),
        # Flips only cluster 1 in first 2 steps.
        (b'near1',
         client_datasets.ClientDataset({'x': np.array([0.9, 1.1, 1.3])}),
         jax.random.PRNGKey(106)),
    ]

    cluster_losses = hyp_cluster._cluster_losses(
        evaluator=evaluator,
        cluster_params=cluster_params,
        clients=clients,
        batch_hparams=batch_hparams)
    self.assertCountEqual(cluster_losses,
                          [b'near0', b'near-1', b'near2', b'near1'])
    npt.assert_allclose(cluster_losses[b'near0'],
                        np.array([0.1, 1.1 + 0.01, 1.9 + 0.02]))
    npt.assert_allclose(cluster_losses[b'near-1'],
                        np.array([1.1, 0.5 / 3 + 0.01, 3.1 + 0.02]))
    npt.assert_allclose(
        cluster_losses[b'near2'],
        np.array([6.1 / 3, 9.1 / 3 + 0.01, 0.1 + 0.02]),
        rtol=1e-6)
    npt.assert_allclose(cluster_losses[b'near1'],
                        np.array([1.1, 0.5 / 3 + 0.01, 0.9 + 0.02]))

    client_cluster_ids = hyp_cluster.maximization_step(
        evaluator=evaluator,
        cluster_params=cluster_params,
        clients=clients,
        batch_hparams=batch_hparams)
    self.assertDictEqual(client_cluster_ids, {
        b'near0': 0,
        b'near-1': 1,
        b'near2': 2,
        b'near1': 1
    })

  def test_hyp_cluster_evaluator(self):
    functions_called = set()

    def apply_for_eval(params, batch):
      functions_called.add('apply_for_eval')
      score = params * batch['x']
      return jnp.stack([-score, score], axis=-1)

    def apply_for_train(params, batch, rng):
      functions_called.add('apply_for_train')
      self.assertIsNotNone(rng)
      return params * batch['x']

    def train_loss(batch, out):
      functions_called.add('train_loss')
      return jnp.abs(batch['y'] * 2 - 1 - out)

    def regularizer(params):
      # Just to check regularizer is called.
      del params
      functions_called.add('regularizer')
      return 0

    evaluator = hyp_cluster.HypClusterEvaluator(
        models.Model(
            init=None,
            apply_for_eval=apply_for_eval,
            apply_for_train=apply_for_train,
            train_loss=train_loss,
            eval_metrics={'accuracy': metrics.Accuracy()}), regularizer)

    cluster_params = [jnp.array(1.), jnp.array(-1.)]
    clients = [
        # Evaluated using cluster 0.
        (b'0',
         client_datasets.ClientDataset({
             'x': np.array([0.9, -0.9, 0.8, -0.8, -0.3]),
             'y': np.array([1, 0, 1, 0, 1])
         }), jax.random.PRNGKey(0)),
        # Evaluated using cluster 1.
        (b'1',
         client_datasets.ClientDataset({
             'x': np.array([3., 2., 1.]),
             'y': np.array([0, 0, 1])
         }), jax.random.PRNGKey(1)),
    ]
    for batch_size in [1, 2, 4]:
      with self.subTest(f'batch_size = {batch_size}'):
        batch_hparams = client_datasets.PaddedBatchHParams(
            batch_size=batch_size)
        metric_values = dict(
            evaluator.evaluate_clients(cluster_params, clients, batch_hparams))
        self.assertCountEqual(metric_values, [b'0', b'1'])
        self.assertCountEqual(metric_values[b'0'], ['accuracy'])
        npt.assert_allclose(metric_values[b'0']['accuracy'], 4 / 5)
        self.assertCountEqual(metric_values[b'1'], ['accuracy'])
        npt.assert_allclose(metric_values[b'1']['accuracy'], 2 / 3)
    self.assertCountEqual(
        functions_called,
        ['apply_for_train', 'train_loss', 'apply_for_eval', 'regularizer'])

  def test_expectation_step(self):

    def per_example_loss(params, batch, rng):
      self.assertIsNotNone(rng)
      return jnp.square(params - batch['x'])

    trainer = hyp_cluster.ClientDeltaTrainer(
        models.grad(per_example_loss), optimizers.sgd(0.5))
    batch_hparams = client_datasets.ShuffleRepeatBatchHParams(
        batch_size=1, num_epochs=5)

    cluster_params = [jnp.array(1.), jnp.array(-1.), jnp.array(3.14)]
    client_cluster_ids = {b'0': 0, b'1': 0, b'2': 1, b'3': 1, b'4': 0}
    # RNGs are not actually used.
    clients = [
        (b'0', client_datasets.ClientDataset({'x': np.array([1.1])}),
         jax.random.PRNGKey(0)),
        (b'1', client_datasets.ClientDataset({'x': np.array([0.9, 0.9])}),
         jax.random.PRNGKey(1)),
        (b'2', client_datasets.ClientDataset({'x': np.array([-1.1])}),
         jax.random.PRNGKey(2)),
        (b'3',
         client_datasets.ClientDataset({'x': np.array([-0.9, -0.9, -0.9])}),
         jax.random.PRNGKey(3)),
        (b'4', client_datasets.ClientDataset({'x': np.array([-0.1])}),
         jax.random.PRNGKey(4)),
    ]
    cluster_delta_params = hyp_cluster.expectation_step(
        trainer=trainer,
        cluster_params=cluster_params,
        client_cluster_ids=client_cluster_ids,
        clients=clients,
        batch_hparams=batch_hparams)
    self.assertIsInstance(cluster_delta_params, list)
    self.assertLen(cluster_delta_params, 3)
    npt.assert_allclose(cluster_delta_params[0], (-0.1 + 0.1 * 2 + 1.1) / 4)
    npt.assert_allclose(cluster_delta_params[1], (0.1 - 0.1 * 3) / 4, rtol=1e-6)
    self.assertIsNone(cluster_delta_params[2])

  def test_hyp_cluster(self):
    functions_called = set()

    def per_example_loss(params, batch, rng):
      self.assertIsNotNone(rng)
      functions_called.add('per_example_loss')
      return jnp.square(params - batch['x'])

    def regularizer(params):
      del params
      functions_called.add('regularizer')
      return 0

    client_optimizer = optimizers.sgd(0.5)
    server_optimizer = optimizers.sgd(0.25)

    maximization_batch_hparams = client_datasets.PaddedBatchHParams(
        batch_size=2)
    expectation_batch_hparams = client_datasets.ShuffleRepeatBatchHParams(
        batch_size=1, num_epochs=5)

    algorithm = hyp_cluster.hyp_cluster(
        per_example_loss=per_example_loss,
        client_optimizer=client_optimizer,
        server_optimizer=server_optimizer,
        maximization_batch_hparams=maximization_batch_hparams,
        expectation_batch_hparams=expectation_batch_hparams,
        regularizer=regularizer)

    init_state = algorithm.init([jnp.array(1.), jnp.array(-1.)])
    # Nothing happens with empty data.
    no_op_state, diagnostics = algorithm.apply(init_state, clients=[])
    npt.assert_array_equal(init_state.cluster_params,
                           no_op_state.cluster_params)
    self.assertEmpty(diagnostics)
    # Some actual training. PRNGKeys are not actually used.
    clients = [
        (b'0', client_datasets.ClientDataset({'x': np.array([1.1])}),
         jax.random.PRNGKey(0)),
        (b'1', client_datasets.ClientDataset({'x': np.array([0.9, 0.9])}),
         jax.random.PRNGKey(1)),
        (b'2', client_datasets.ClientDataset({'x': np.array([-1.1])}),
         jax.random.PRNGKey(2)),
        (b'3',
         client_datasets.ClientDataset({'x': np.array([-0.9, -0.9, -0.9])}),
         jax.random.PRNGKey(3)),
    ]
    next_state, diagnostics = algorithm.apply(init_state, clients)
    npt.assert_equal(
        diagnostics, {
            b'0': {
                'cluster_id': 0
            },
            b'1': {
                'cluster_id': 0
            },
            b'2': {
                'cluster_id': 1
            },
            b'3': {
                'cluster_id': 1
            },
        })
    cluster_params = next_state.cluster_params
    self.assertIsInstance(cluster_params, list)
    self.assertLen(cluster_params, 2)
    npt.assert_allclose(cluster_params[0], [1. - 0.25 * 0.1 / 3])
    npt.assert_allclose(cluster_params[1], [-1. + 0.25 * 0.2 / 4])
    self.assertCountEqual(functions_called, ['per_example_loss', 'regularizer'])


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Metrics."""

import abc
import numbers
from typing import Any, Optional, Tuple

from fedjax.legacy.core import dataclasses
import jax
import jax.numpy as jnp

# Small constant to add to denominator to avoid division by 0.
_SAFE_DIVIDE = 1e-10


class Metric(metaclass=abc.ABCMeta):
  """Interface for all metric containers (e.g. accuracy).

  `Metric` stores intermediate values as well as methods for accumulation and
  final result computation.
  """

  @abc.abstractmethod
  def merge(self, other: 'Metric') -> 'Metric':
    """Merges `self` and `other` into a new single accumulated metric."""

  @abc.abstractmethod
  def result(self) -> jnp.ndarray:
    """Computes final metric result from intermediate values."""

  def __str__(self) -> str:
    """Returns human readable string representation of metric."""
    return f'{repr(self)} => {self.result()}'


def _is_scalar(x):
  if isinstance(x, jnp.ndarray):
    return x.ndim == 0
  return isinstance(x, numbers.Number)


@dataclasses.dataclass
class MeanMetric(Metric):
  """Implementation for metrics that are reduced by averaging (total / count).

  Attributes:
    total: Scalar sum of intermediate values.
    count: Scalar number of intermediate values.
  """

  total: jnp.ndarray
  count: jnp.ndarray

  def __post_init__(self):
    if not (_is_scalar(self.total) and _is_scalar(self.count)):
      raise TypeError('total and count must both be scalars.')

  @classmethod
  def from_values(cls,
                  values: jnp.ndarray,
                  weights: Optional[jnp.ndarray] = None) -> 'MeanMetric':
    """Constructs MeanMetric from intermediate values and optional weights.

    Args:
      values: Array of intermediate values.
      weights: Array of weights for each intermediate value of the same shape as
        values. Defaults to unweighted.

    Returns:
      MeanMetric for (possibly weighted) average of values.
    """
    if weights is None:
      weights = jnp.ones_like(values)
    return cls(total=jnp.sum(values * weights), count=jnp.sum(weights))

  def merge(self, other: 'MeanMetric') -> 'MeanMetric':
    return type(self)(
        total=self.total + other.total, count=self.count + other.count)

  def result(self) -> jnp.ndarray:
    return self.total / jnp.maximum(self.count, _SAFE_DIVIDE)


@dataclasses.dataclass
class CountMetric(Metric):
  """Implementation for counter metrics (e.g. num_out_of_vocabulary_words)."""

  count: jnp.ndarray

  def __post_init__(self):
    if not _is_scalar(self.count):
      raise TypeError('count must be a scalar.')

  def merge(self, other: 'CountMetric') -> 'CountMetric':
    return type(self)(count=self.count + other.count)

  def result(self) -> jnp.ndarray:
    return self.count


def _unreduced_cross_entropy_loss_fn(targets: jnp.ndarray,
                                     preds: jnp.ndarray) -> jnp.ndarray:
  """Returns unreduced cross entropy loss."""
  num_classes = preds.shape[-1]
  log_preds = jax.nn.log_softmax(preds)
  one_hot_targets = jax.nn.one_hot(targets, num_classes)
  return -jnp.sum(one_hot_targets * log_preds, axis=-1)


def cross_entropy_loss_fn(targets: jnp.ndarray,
                          preds: jnp.ndarray) -> MeanMetric:
  """Computes cross entropy loss.

  Args:
    targets: Target values of shape [batch_size, ...] in range [0, num_classes).
    preds: Unnormalized model output of shape [batch_size, ..., num_classes].

  Returns:
    Metric for loss.
  """
  unreduced_loss = _unreduced_cross_entropy_loss_fn(targets, preds)
  return MeanMetric.from_values(unreduced_loss)


def masked_cross_entropy_loss_fn(
    targets: jnp.ndarray, preds: jnp.ndarray,
    mask_values: Tuple[int, ...] = ()) -> MeanMetric:
  """Computes cross entropy loss after discounting masked values.

  Args:
    targets: Target values of shape [batch_size, ...] in range [0, num_classes).
    preds: Unnormalized model output of shape [batch_size, ..., num_classes].
    mask_values: Target values to be masked and not counted in loss.

  Returns:
    Metric for masked loss.
  """
  weights = jnp.ones_like(targets, dtype=preds.dtype)
  for mv in mask_values:
    weights *= (targets != mv)
  unreduced_loss = _unreduced_cross_entropy_loss_fn(targets, preds)
  return MeanMetric.from_values(unreduced_loss, weights=weights)


def accuracy_fn(targets: jnp.ndarray, preds: jnp.ndarray) -> MeanMetric:
  """Computes accuracy.

  Args:
    targets: Target values of shape [batch_size, ...] in range [0, num_classes).
    preds: Unnormalized model output of shape [batch_size, ..., num_classes].

  Returns:
    Metric for accuracy.
  """
  pred_class = jnp.argmax(preds, axis=-1)
  return MeanMetric.from_values(pred_class == targets)


def masked_accuracy_fn(
    targets: jnp.ndarray,
    preds: jnp.ndarray,
    mask_values: Tuple[int, ...] = (),) -> MeanMetric:
  """Computes accuracy after discounting masked values.

  Args:
    targets: Target values of shape [batch_size, ...] in range [0, num_classes).
    preds: Unnormalized model output of shape [batch_size, ..., num_classes].
    mask_values: Target values to be masked and not counted in accuracy.

  Returns:
    Metric for masked accuracy.
  """
  weights = jnp.ones_like(targets, dtype=preds.dtype)
  for mv in mask_values:
    weights *= (targets != mv)
  pred_class = jnp.argmax(preds, axis=-1)
  return MeanMetric.from_values(pred_class == targets, weights=weights)


def masked_accuracy_fn_with_logits_mask(
    targets: jnp.ndarray,
    preds: jnp.ndarray,
    logits_mask: jnp.ndarray,
    mask_values: Tuple[int, ...] = (),
) -> MeanMetric:
  """Computes accuracy after discounting masked values.

  Args:
    targets: Target values of shape [batch_size, ...] in range [0, num_classes).
    preds: Unnormalized model output of shape [batch_size, ..., num_classes].
    logits_mask: Mask of shape [num_classes] to be applied for preds.
    mask_values: Target values to be masked and not counted in accuracy.

  Returns:
    Metric for masked accuracy with logits mask.
  """
  weights = jnp.ones_like(targets, dtype=preds.dtype)
  for mv in mask_values:
    weights *= (targets != mv)
  preds = preds + logits_mask
  pred_class = jnp.argmax(preds, axis=-1)
  return MeanMetric.from_values(pred_class == targets, weights=weights)


def masked_count(targets: jnp.ndarray,
                 mask_values: Tuple[Any, ...] = ()) -> CountMetric:
  """Counts total number of non masked targets."""
  weights = jnp.ones_like(targets, dtype=jnp.int32)
  for mv in mask_values:
    weights *= (targets != mv)
  return CountMetric(count=jnp.sum(weights))


def truncation_rate(targets: jnp.ndarray,
                    eos_value: int,
                    pad_value: int) -> MeanMetric:
  """Computes the proportion of sequence examples that were truncated.

  Args:
    targets: Target values of shape [batch_size, sequence_length, ...].
    eos_value: Target value denoting end of sequence. Truncated sequences will
      not have this value.
    pad_value: Optional target value for padding to discount empty sequences.

  Returns:
    Metric for trucation rate.
  """
  not_empty = jnp.any(targets != pad_value, axis=1)
  is_truncated = jnp.all(targets != eos_value, axis=1) * not_empty
  return MeanMetric(total=jnp.sum(is_truncated), count=jnp.sum(not_empty))


def oov_rate(
    targets: jnp.ndarray,
    oov_values: Tuple[int, ...],
    mask_values: Tuple[int, ...] = ()) -> MeanMetric:
  """Computes proportion of non masked tokens that are out of vocabulary.

  Args:
    targets: Target values of shape [batch_size, sequence_length, ...].
    oov_values: Target values denoting out of vocabulary values.
    mask_values: Target values to be masked and not counted in metric.

  Returns:
    Metric for out of vocabulary rate.
  """
  weights = jnp.ones_like(targets, dtype=jnp.float32)
  for mv in mask_values:
    weights *= (targets != mv)
  num_non_masked = jnp.sum(weights)
  for ov in oov_values:
    weights *= (targets == ov)
  num_oov = jnp.sum(weights)
  return MeanMetric(total=num_oov, count=num_non_masked)


def sequence_length(targets: jnp.ndarray, pad_value: int) -> MeanMetric:
  """Computes length of sequence examples by number of non-pad tokens."""
  non_pad_mask = targets != pad_value
  not_empty = jnp.any(non_pad_mask, axis=1)
  num_non_pad = jnp.sum(non_pad_mask, axis=1)
  return MeanMetric(total=jnp.sum(num_non_pad), count=jnp.sum(not_empty))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.models.emnist."""

from absl.testing import absltest

from fedjax.core import tree_util
from fedjax.models import emnist

import jax
import jax.numpy as jnp


class EmnistModelTest(absltest.TestCase):

  def check_model(self, model):
    rng = jax.random.PRNGKey(0)
    params = model.init(rng)
    batch = {'x': jnp.ones((5, 28, 28, 1)), 'y': jnp.ones((5,))}
    with self.subTest('apply_for_train'):
      preds = model.apply_for_train(params, batch, rng)
      self.assertTupleEqual(preds.shape, (5, 62))
    with self.subTest('apply_for_eval'):
      preds = model.apply_for_eval(params, batch)
      self.assertTupleEqual(preds.shape, (5, 62))
    with self.subTest('train_loss'):
      preds = model.apply_for_train(params, batch, rng)
      train_loss = model.train_loss(batch, preds)
      self.assertTupleEqual(train_loss.shape, (5,))

  def test_create_conv_model(self):
    model = emnist.create_conv_model(only_digits=False)
    params = model.init(jax.random.PRNGKey(0))
    self.assertEqual(tree_util.tree_size(params), 1206590)
    self.check_model(model)

  def test_create_dense_model(self):
    model = emnist.create_dense_model(only_digits=False, hidden_units=200)
    params = model.init(jax.random.PRNGKey(0))
    self.assertEqual(tree_util.tree_size(params), 209662)
    self.check_model(model)

  def test_create_logistic_model(self):
    model = emnist.create_logistic_model(only_digits=False)
    params = model.init(jax.random.PRNGKey(0))
    self.assertEqual(tree_util.tree_size(params), 48670)
    self.check_model(model)

  def test_create_stax_dense_model(self):
    model = emnist.create_stax_dense_model(only_digits=False, hidden_units=200)
    params = model.init(jax.random.PRNGKey(0))
    self.assertEqual(tree_util.tree_size(params), 209662)
    self.check_model(model)

if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.algorithms.mime_lite."""

from fedjax.legacy import core
from fedjax.legacy.algorithms import mime
from fedjax.legacy.algorithms import mime_lite
import tensorflow as tf


class Mimetest(tf.test.TestCase):

  def test_run(self):
    num_classes = 10
    data, model = core.test_util.create_toy_example(
        num_clients=10,
        num_clusters=4,
        num_classes=num_classes,
        num_examples=5,
        seed=0)
    rng_seq = core.PRNGSequence(0)
    algorithm = mime.Mime(
        federated_data=data,
        model=model,
        base_optimizer=core.get_optimizer(
            core.OptimizerName.MOMENTUM, learning_rate=0.2, momentum=0.9),
        hparams=mime_lite.MimeLiteHParams(
            train_data_hparams=core.ClientDataHParams(batch_size=100),
            combined_data_hparams=core.ClientDataHParams(batch_size=100),
            server_learning_rate=1.0),
        rng_seq=rng_seq,
    )
    dataset = core.create_tf_dataset_for_clients(data).batch(50)

    state = algorithm.init_state()
    init_metrics = core.evaluate_single_client(
        dataset=dataset, params=state.params, model=model)
    for _ in range(10):
      state = algorithm.run_round(state, data.client_ids)
    metrics = core.evaluate_single_client(
        dataset=dataset, params=state.params, model=model)

    self.assertLess(metrics['loss'], init_metrics['loss'])
    self.assertGreater(metrics['accuracy'], init_metrics['accuracy'])


if __name__ == '__main__':
  tf.test.main()
